import json
import os
import warnings
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast,
    final,
)

import cloudpickle
import numpy as np
import pandas as pd
from packaging import version
from typing_extensions import TypeGuard, Unpack

from snowflake.ml._internal import type_utils
from snowflake.ml.model import custom_model, model_signature, type_hints as model_types
from snowflake.ml.model._packager.model_env import model_env
from snowflake.ml.model._packager.model_handlers import _base, _utils as handlers_utils
from snowflake.ml.model._packager.model_handlers_migrator import base_migrator
from snowflake.ml.model._packager.model_meta import (
    model_blob_meta,
    model_meta as model_meta_api,
    model_meta_schema,
)
from snowflake.ml.model._signatures import (
    builtins_handler,
    utils as model_signature_utils,
)
from snowflake.ml.model.models import huggingface_pipeline
from snowflake.snowpark._internal import utils as snowpark_utils

if TYPE_CHECKING:
    import transformers


def get_requirements_from_task(task: str, spcs_only: bool = False) -> List[model_env.ModelDependency]:
    # Text
    if task in [
        "conversational",
        "fill-mask",
        "ner",
        "token-classification",
        "question-answering",
        "summarization",
        "table-question-answering",
        "text-classification",
        "sentiment-analysis",
        "text-generation",
        "text2text-generation",
        "zero-shot-classification",
    ] or task.startswith("translation"):
        return (
            [model_env.ModelDependency(requirement="tokenizers>=0.13.3", pip_name="tokenizers")]
            if spcs_only
            else [model_env.ModelDependency(requirement="tokenizers", pip_name="tokenizers")]
        )

    return []


class NumpyEncoder(json.JSONEncoder):
    # This is a JSON encoder class to ensure the output from Huggingface pipeline is JSON serializable.
    # What it covers is numpy object.
    def default(self, z: object) -> object:
        if isinstance(z, np.number):
            if np.can_cast(z, np.int64, casting="safe"):
                return int(z)
            elif np.can_cast(z, np.float64, casting="safe"):
                return z.astype(np.float64)
        return super().default(z)


@final
class HuggingFacePipelineHandler(
    _base.BaseModelHandler[Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"]]
):
    """Handler for custom model."""

    HANDLER_TYPE = "huggingface_pipeline"
    HANDLER_VERSION = "2023-12-01"
    _MIN_SNOWPARK_ML_VERSION = "1.0.12"
    _HANDLER_MIGRATOR_PLANS: Dict[str, Type[base_migrator.BaseModelHandlerMigrator]] = {}

    MODELE_BLOB_FILE_OR_DIR = "model"
    ADDITIONAL_CONFIG_FILE = "pipeline_config.pt"
    DEFAULT_TARGET_METHODS = ["__call__"]
    IS_AUTO_SIGNATURE = True

    @classmethod
    def can_handle(
        cls,
        model: model_types.SupportedModelType,
    ) -> TypeGuard[Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"]]:
        if type_utils.LazyType("transformers.Pipeline").isinstance(model):
            return True
        if isinstance(model, huggingface_pipeline.HuggingFacePipelineModel):
            return True
        return False

    @classmethod
    def cast_model(
        cls,
        model: model_types.SupportedModelType,
    ) -> Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"]:
        try:
            if isinstance(model, huggingface_pipeline.HuggingFacePipelineModel):
                raise ImportError
            else:
                import transformers
        except ImportError:
            assert isinstance(model, huggingface_pipeline.HuggingFacePipelineModel)
            return model
        else:
            assert isinstance(model, transformers.Pipeline)
            return model

    @classmethod
    def save_model(
        cls,
        name: str,
        model: Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"],
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        sample_input: Optional[model_types.SupportedDataType] = None,
        is_sub_model: Optional[bool] = False,
        **kwargs: Unpack[model_types.HuggingFaceSaveOptions],
    ) -> None:
        if type_utils.LazyType("transformers.Pipeline").isinstance(model):
            task = model.task  # type:ignore[attr-defined]
            framework = model.framework  # type:ignore[attr-defined]
            batch_size = model._batch_size  # type:ignore[attr-defined]
        else:
            assert isinstance(model, huggingface_pipeline.HuggingFacePipelineModel)
            task = model.task
            framework = getattr(model, "framework", None)
            batch_size = getattr(model, "batch_size", None)

        if type_utils.LazyType("transformers.Pipeline").isinstance(model):
            params = {
                **model._preprocess_params,  # type:ignore[attr-defined]
                **model._forward_params,  # type:ignore[attr-defined]
                **model._postprocess_params,  # type:ignore[attr-defined]
            }
        else:
            assert isinstance(model, huggingface_pipeline.HuggingFacePipelineModel)
            params = {**model.__dict__, **model.model_kwargs}

        inferred_pipe_sig = model_signature_utils.huggingface_pipeline_signature_auto_infer(task, params=params)

        if not is_sub_model:
            target_methods = handlers_utils.get_target_methods(
                model=model,
                target_methods=kwargs.pop("target_methods", None),
                default_target_methods=cls.DEFAULT_TARGET_METHODS,
            )

            if model_meta.signatures:
                if type_utils.LazyType("transformers.Pipeline").isinstance(model):
                    handlers_utils.validate_target_methods(model, list(model_meta.signatures.keys()))
                else:
                    warnings.warn(
                        "It is impossible to validate your model signatures when using a"
                        " `snowflake.ml.model.models.huggingface_pipeline.HuggingFacePipelineModel` object. "
                        "Please make sure you are providing correct model signatures.",
                        UserWarning,
                    )
            else:
                handlers_utils.validate_target_methods(model, target_methods)
                if sample_input is not None:
                    warnings.warn(
                        "Inferring model signature from sample input for hugggingface pipeline is not supported. "
                        + "Model signature will automatically be inferred from pipeline task. "
                        + "Or, you could specify model signature manually.",
                        UserWarning,
                    )
                if inferred_pipe_sig is None:
                    raise NotImplementedError(f"Cannot auto infer the signature of pipeline for task {task}")

                model_meta.signatures = {"__call__": inferred_pipe_sig}

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        os.makedirs(model_blob_path, exist_ok=True)

        if type_utils.LazyType("transformers.Pipeline").isinstance(model):
            model.save_pretrained(  # type:ignore[attr-defined]
                os.path.join(model_blob_path, cls.MODELE_BLOB_FILE_OR_DIR)
            )
            pipeline_params = {
                "_batch_size": model._batch_size,  # type:ignore[attr-defined]
                "_num_workers": model._num_workers,  # type:ignore[attr-defined]
                "_preprocess_params": model._preprocess_params,  # type:ignore[attr-defined]
                "_forward_params": model._forward_params,  # type:ignore[attr-defined]
                "_postprocess_params": model._postprocess_params,  # type:ignore[attr-defined]
            }
            with open(
                os.path.join(
                    model_blob_path,
                    cls.MODELE_BLOB_FILE_OR_DIR,
                    cls.ADDITIONAL_CONFIG_FILE,
                ),
                "wb",
            ) as f:
                cloudpickle.dump(pipeline_params, f)
        else:
            with open(
                os.path.join(model_blob_path, cls.MODELE_BLOB_FILE_OR_DIR),
                "wb",
            ) as f:
                cloudpickle.dump(model, f)

        base_meta = model_blob_meta.ModelBlobMeta(
            name=name,
            model_type=cls.HANDLER_TYPE,
            handler_version=cls.HANDLER_VERSION,
            path=cls.MODELE_BLOB_FILE_OR_DIR,
            options=model_meta_schema.HuggingFacePipelineModelBlobOptions(
                {
                    "task": task,
                    "batch_size": batch_size if batch_size is not None else 1,
                }
            ),
        )
        model_meta.models[name] = base_meta
        model_meta.min_snowpark_ml_version = cls._MIN_SNOWPARK_ML_VERSION

        pkgs_requirements = [
            model_env.ModelDependency(requirement="transformers>=4.32.1", pip_name="transformers"),
        ] + get_requirements_from_task(
            task, spcs_only=(not type_utils.LazyType("transformers.Pipeline").isinstance(model))
        )
        if framework is None or framework == "pt":
            # Since we set default cuda version to be 11.7, to make sure it works with GPU, we need to have a default
            # Pytorch version that works with CUDA 11.7 as well. This is required for huggingface pipelines only as
            # users are not required to install pytorch locally if they are using the wrapper.
            pkgs_requirements.append(model_env.ModelDependency(requirement="pytorch==2.0.1", pip_name="torch"))
        elif framework == "tf":
            pkgs_requirements.append(model_env.ModelDependency(requirement="tensorflow", pip_name="tensorflow"))
        model_meta.env.include_if_absent(
            pkgs_requirements, check_local_version=(type_utils.LazyType("transformers.Pipeline").isinstance(model))
        )
        model_meta.env.cuda_version = kwargs.get("cuda_version", model_env.DEFAULT_CUDA_VERSION)

    @staticmethod
    def _get_device_config() -> Dict[str, str]:
        device_config = {}
        device_config["device_map"] = "auto"

        return device_config

    @classmethod
    def load_model(
        cls,
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"]:
        if snowpark_utils.is_in_stored_procedure():  # type: ignore[no-untyped-call]
            # We need to redirect the some folders to a writable location in the sandbox.
            os.environ["HF_HOME"] = "/tmp"
            os.environ["XDG_CACHE_HOME"] = "/tmp"

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        model_blobs_metadata = model_meta.models
        model_blob_metadata = model_blobs_metadata[name]
        model_blob_filename = model_blob_metadata.path
        model_blob_options = cast(model_meta_schema.HuggingFacePipelineModelBlobOptions, model_blob_metadata.options)
        if "task" not in model_blob_options:
            raise ValueError("Missing field `task` in model blob metadata for type `huggingface_pipeline`")
        if "batch_size" not in model_blob_options:
            raise ValueError("Missing field `batch_size` in model blob metadata for type `huggingface_pipeline`")

        model_blob_file_or_dir_path = os.path.join(model_blob_path, model_blob_filename)
        if os.path.isdir(model_blob_file_or_dir_path):
            import transformers

            with open(
                os.path.join(
                    model_blob_file_or_dir_path,
                    cls.ADDITIONAL_CONFIG_FILE,
                ),
                "rb",
            ) as f:
                pipeline_params = cloudpickle.load(f)

            if kwargs.get("use_gpu", False):
                device_config = cls._get_device_config()
            else:
                device_config = {}

            m = transformers.pipeline(
                model_blob_options["task"],
                model=model_blob_file_or_dir_path,
                **device_config,
            )

            m.__dict__.update(pipeline_params)

        else:
            assert os.path.isfile(model_blob_file_or_dir_path)
            with open(model_blob_file_or_dir_path, "rb") as f:
                m = cloudpickle.load(f)
            assert isinstance(m, huggingface_pipeline.HuggingFacePipelineModel)
            if (
                getattr(m, "device", None) is None
                and getattr(m, "device_map", None) is None
                and kwargs.get("use_gpu", False)
            ):
                m.__dict__.update(cls._get_device_config())

            if getattr(m, "torch_dtype", None) is None and kwargs.get("use_gpu", False):
                m.__dict__.update(torch_dtype="auto")
        return m

    @classmethod
    def convert_as_custom_model(
        cls,
        raw_model: Union[huggingface_pipeline.HuggingFacePipelineModel, "transformers.Pipeline"],
        model_meta: model_meta_api.ModelMetadata,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> custom_model.CustomModel:
        import transformers

        from snowflake.ml.model import custom_model

        def _create_custom_model(
            raw_model: "transformers.Pipeline",
            model_meta: model_meta_api.ModelMetadata,
        ) -> Type[custom_model.CustomModel]:
            def fn_factory(
                raw_model: "transformers.Pipeline",
                signature: model_signature.ModelSignature,
                target_method: str,
            ) -> Callable[[custom_model.CustomModel, pd.DataFrame], pd.DataFrame]:
                @custom_model.inference_api
                def fn(self: custom_model.CustomModel, X: pd.DataFrame) -> pd.DataFrame:
                    # These 3 zero-shot classification cannot take a list of dict as input like other multi input
                    # pipelines, thus dealing separately.
                    if isinstance(
                        raw_model,
                        (
                            transformers.ZeroShotAudioClassificationPipeline,
                            transformers.ZeroShotClassificationPipeline,
                            transformers.ZeroShotImageClassificationPipeline,
                        ),
                    ):
                        temp_res = X.apply(
                            lambda row: getattr(raw_model, target_method)(
                                row[signature.inputs[0].name], row["candidate_labels"]
                            ),
                            axis=1,
                        ).to_list()
                    else:
                        # For others, we could offer the whole dataframe as a list.
                        # Some of them may need some conversion
                        if isinstance(raw_model, transformers.ConversationalPipeline):
                            input_data = [
                                transformers.Conversation(
                                    text=conv_data["user_inputs"][0],
                                    past_user_inputs=conv_data["user_inputs"][1:],
                                    generated_responses=conv_data["generated_responses"],
                                )
                                for conv_data in X.to_dict("records")
                            ]
                        elif len(signature.inputs) == 1:
                            input_data = X.to_dict("list")[signature.inputs[0].name]
                        else:
                            if isinstance(raw_model, transformers.TableQuestionAnsweringPipeline):
                                X["table"] = X["table"].apply(json.loads)

                            input_data = X.to_dict("records")
                        temp_res = getattr(raw_model, target_method)(input_data)

                    # Some huggingface pipeline will omit the outer list when there is only 1 input.
                    # Making it not aligned with the auto-inferred signature.
                    # If the output is a dict, we could blindly create a list containing that.
                    # Otherwise, creating pandas DataFrame won't succeed.
                    if isinstance(temp_res, (dict, transformers.Conversation)) or (
                        # For some pipeline that is expected to generate a list of dict per input
                        # When it omit outer list, it becomes list of dict instead of list of list of dict.
                        # We need to distinguish them from those pipelines that designed to output a dict per input
                        # So we need to check the pipeline type.
                        isinstance(
                            raw_model,
                            (
                                transformers.FillMaskPipeline,
                                transformers.QuestionAnsweringPipeline,
                            ),
                        )
                        and X.shape[0] == 1
                        and isinstance(temp_res[0], dict)
                    ):
                        temp_res = [temp_res]

                    if len(temp_res) == 0:
                        return pd.DataFrame()

                    if isinstance(raw_model, transformers.ConversationalPipeline):
                        temp_res = [[conv.generated_responses] for conv in temp_res]

                    # To concat those who outputs a list with one input.
                    if builtins_handler.ListOfBuiltinHandler.can_handle(temp_res):
                        res = builtins_handler.ListOfBuiltinHandler.convert_to_df(temp_res)
                    elif isinstance(temp_res[0], dict):
                        res = pd.DataFrame(temp_res)
                    elif isinstance(temp_res[0], list):
                        res = pd.DataFrame([json.dumps(output, cls=NumpyEncoder) for output in temp_res])
                    else:
                        raise ValueError(f"Cannot parse output {temp_res} from pipeline object")

                    return model_signature_utils.rename_pandas_df(data=res, features=signature.outputs)

                return fn

            type_method_dict = {}
            for target_method_name, sig in model_meta.signatures.items():
                type_method_dict[target_method_name] = fn_factory(raw_model, sig, target_method_name)

            _HFPipelineModel = type(
                "_HFPipelineModel",
                (custom_model.CustomModel,),
                type_method_dict,
            )

            return _HFPipelineModel

        if isinstance(raw_model, huggingface_pipeline.HuggingFacePipelineModel):
            if version.parse(transformers.__version__) < version.parse("4.32.0"):
                # Backward compatibility since HF interface change.
                raw_model.__dict__["use_auth_token"] = raw_model.__dict__["token"]
                del raw_model.__dict__["token"]
            pipe = transformers.pipeline(**raw_model.__dict__)
        else:
            pipe = raw_model

        pipe.binary_output = False

        # To enable batch_size > 1 for LLM
        # Pipe might not have tokenizer, but should always have a model, and model should always have a config.
        if getattr(pipe, "tokenizer", None) is not None and pipe.tokenizer.pad_token_id is None:
            pipe.tokenizer.pad_token_id = pipe.model.config.eos_token_id

        _HFPipelineModel = _create_custom_model(pipe, model_meta)
        hg_pipe_model = _HFPipelineModel(custom_model.ModelContext())

        return hg_pipe_model



import warnings
from typing import Any, List, Literal, Optional, Sequence, Tuple, Type

import numpy as np
import pandas as pd

import snowflake.snowpark
import snowflake.snowpark.functions as F
import snowflake.snowpark.types as spt
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import formatting, identifier
from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._signatures import (
    base_handler,
    builtins_handler as builtins_handler,
    core,
    numpy_handler,
    pandas_handler,
    pytorch_handler,
    snowpark_handler,
    tensorflow_handler,
    utils,
)

DataType = core.DataType
BaseFeatureSpec = core.BaseFeatureSpec
FeatureSpec = core.FeatureSpec
FeatureGroupSpec = core.FeatureGroupSpec
ModelSignature = core.ModelSignature


_LOCAL_DATA_HANDLERS: List[Type[base_handler.BaseDataHandler[Any]]] = [
    pandas_handler.PandasDataFrameHandler,
    numpy_handler.NumpyArrayHandler,
    builtins_handler.ListOfBuiltinHandler,
    numpy_handler.SeqOfNumpyArrayHandler,
    pytorch_handler.SeqOfPyTorchTensorHandler,
    tensorflow_handler.SeqOfTensorflowTensorHandler,
]
_ALL_DATA_HANDLERS = _LOCAL_DATA_HANDLERS + [snowpark_handler.SnowparkDataFrameHandler]


def _truncate_data(
    data: model_types.SupportedDataType,
) -> model_types.SupportedDataType:
    for handler in _ALL_DATA_HANDLERS:
        if handler.can_handle(data):
            row_count = handler.count(data)
            if row_count <= handler.SIG_INFER_ROWS_COUNT_LIMIT:
                return data

            warnings.warn(
                formatting.unwrap(
                    f"""
                    The sample input has {row_count} rows, thus a truncation happened before inferring signature.
                    This might cause inaccurate signature inference.
                    If that happens, consider specifying signature manually.
                    """
                ),
                category=UserWarning,
                stacklevel=1,
            )
            return handler.truncate(data)
    raise snowml_exceptions.SnowflakeMLException(
        error_code=error_codes.NOT_IMPLEMENTED,
        original_exception=NotImplementedError(
            f"Unable to infer model signature: Un-supported type provided {type(data)} for data truncate."
        ),
    )


def _infer_signature(
    data: model_types.SupportedLocalDataType, role: Literal["input", "output"], use_snowflake_identifiers: bool = False
) -> Sequence[core.BaseFeatureSpec]:
    """Infer the inputs/outputs signature given a data that could be dataframe, numpy array or list.
        Dispatching is used to separate logic for different types.
        (Not using Python's singledispatch for unsupported feature of union dispatching in 3.8)

    Args:
        data: The data that we want to infer signature from.
        role: a flag indicating that if this is to infer an input or output feature.
        use_snowflake_identifiers: a flag indicating whether to ensure the signature names are
            valid snowflake identifiers.

    Raises:
        SnowflakeMLException: NotImplementedError: Raised when an unsupported data type is provided.

    Returns:
        A sequence of feature specifications and feature group specifications.
    """
    signature = None
    for handler in _ALL_DATA_HANDLERS:
        if handler.can_handle(data):
            handler.validate(data)
            signature = handler.infer_signature(data, role)
            break

    if signature is None:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.NOT_IMPLEMENTED,
            original_exception=NotImplementedError(
                f"Unable to infer model signature: Un-supported type provided {type(data)} for X type inference."
            ),
        )

    if use_snowflake_identifiers:
        signature = _rename_signature_with_snowflake_identifiers(signature)

    return signature


def _rename_signature_with_snowflake_identifiers(
    signature: Sequence[core.BaseFeatureSpec],
) -> Sequence[core.BaseFeatureSpec]:
    inferred_names = []
    for feature_spec in signature:
        name = identifier.rename_to_valid_snowflake_identifier(feature_spec.name)
        inferred_names.append(name)

    signature = utils.rename_features(signature, inferred_names)

    return signature


def _validate_numpy_array(arr: model_types._SupportedNumpyArray, feature_type: core.DataType) -> bool:
    if feature_type in [
        core.DataType.INT8,
        core.DataType.INT16,
        core.DataType.INT32,
        core.DataType.INT64,
        core.DataType.UINT8,
        core.DataType.UINT16,
        core.DataType.UINT32,
        core.DataType.UINT64,
    ]:
        if not (np.issubdtype(arr.dtype, np.integer)):
            return False
        min_v, max_v = arr.min(), arr.max()
        return bool(max_v <= np.iinfo(feature_type._numpy_type).max and min_v >= np.iinfo(feature_type._numpy_type).min)
    elif feature_type in [core.DataType.FLOAT, core.DataType.DOUBLE]:
        if not (np.issubdtype(arr.dtype, np.integer) or np.issubdtype(arr.dtype, np.floating)):
            return False
        min_v, max_v = arr.min(), arr.max()
        return bool(
            max_v <= np.finfo(feature_type._numpy_type).max  # type: ignore[arg-type]
            and min_v >= np.finfo(feature_type._numpy_type).min  # type: ignore[arg-type]
        )
    else:
        return np.can_cast(arr.dtype, feature_type._numpy_type, casting="no")


def _validate_pandas_df(data: pd.DataFrame, features: Sequence[core.BaseFeatureSpec]) -> None:
    """It validates pandas dataframe with provided features.

    Args:
        data: A pandas dataframe to be validated.
        features: A sequence of feature specifications and feature group specifications, where the dataframe should fit.

    Raises:
        SnowflakeMLException: NotImplementedError: FeatureGroupSpec is not supported.
        SnowflakeMLException: ValueError: Raised when a feature cannot be found.
        SnowflakeMLException: ValueError: Raised when feature is scalar but confront list element.
        SnowflakeMLException: ValueError: Raised when feature type is not aligned in list element.
        SnowflakeMLException: ValueError: Raised when feature shape is not aligned in list element.
        SnowflakeMLException: ValueError: Raised when feature is scalar but confront array element.
        SnowflakeMLException: ValueError: Raised when feature type is not aligned in numpy array element.
        SnowflakeMLException: ValueError: Raised when feature shape is not aligned in numpy array element.
        SnowflakeMLException: ValueError: Raised when feature type is not aligned in string element.
        SnowflakeMLException: ValueError: Raised when feature type is not aligned in bytes element.
    """
    for feature in features:
        ft_name = feature.name
        try:
            data_col = data[ft_name]
        except KeyError:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(f"Data Validation Error: feature {ft_name} does not exist in data."),
            )

        df_col_dtype = data_col.dtype
        if isinstance(feature, core.FeatureGroupSpec):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_IMPLEMENTED,
                original_exception=NotImplementedError("FeatureGroupSpec is not supported."),
            )

        assert isinstance(feature, core.FeatureSpec)  # assert for mypy.
        ft_type = feature._dtype
        ft_shape = feature._shape
        if df_col_dtype != np.dtype("O"):
            if not _validate_numpy_array(data_col.to_numpy(), ft_type):
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError(
                        f"Data Validation Error in feature {ft_name}: "
                        + f"Feature type {ft_type} is not met by all elements in {data_col}."
                    ),
                )
            elif ft_shape is not None:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_DATA,
                    original_exception=ValueError(
                        f"Data Validation Error in feature {ft_name}: "
                        + "Feature is a array type feature while scalar data is provided."
                    ),
                )
        else:
            if isinstance(data_col[0], list):
                if not ft_shape:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + "Feature is a scalar feature while list data is provided."
                        ),
                    )

                converted_data_list = [utils.convert_list_to_ndarray(data_row) for data_row in data_col]

                if not all(_validate_numpy_array(converted_data, ft_type) for converted_data in converted_data_list):
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + f"Feature type {ft_type} is not met by all elements in {data_col}."
                        ),
                    )

                if ft_shape and ft_shape != (-1,):
                    if not all(np.shape(converted_data) == ft_shape for converted_data in converted_data_list):
                        raise snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.INVALID_DATA,
                            original_exception=ValueError(
                                f"Data Validation Error in feature {ft_name}: "
                                + f"Feature shape {ft_shape} is not met by all elements in {data_col}."
                            ),
                        )

            elif isinstance(data_col[0], np.ndarray):
                if not ft_shape:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + "Feature is a scalar feature while array data is provided."
                        ),
                    )

                if not all(_validate_numpy_array(data_row, ft_type) for data_row in data_col):
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + f"Feature type {ft_type} is not met by all elements in {data_col}."
                        ),
                    )

                ft_shape = feature._shape
                if ft_shape and ft_shape != (-1,):
                    if not all(np.shape(data_row) == ft_shape for data_row in data_col):
                        ft_shape = (-1,)
                        raise snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.INVALID_DATA,
                            original_exception=ValueError(
                                f"Data Validation Error in feature {ft_name}: "
                                + f"Feature shape {ft_shape} is not met by all elements in {data_col}."
                            ),
                        )

            elif isinstance(data_col[0], str):
                if ft_shape is not None:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + "Feature is a array type feature while scalar data is provided."
                        ),
                    )

                if ft_type != core.DataType.STRING:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + f"Feature type {ft_type} is not met by all elements in {data_col}."
                        ),
                    )

            elif isinstance(data_col[0], bytes):
                if ft_shape is not None:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + "Feature is a array type feature while scalar data is provided."
                        ),
                    )

                if ft_type != core.DataType.BYTES:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INVALID_DATA,
                        original_exception=ValueError(
                            f"Data Validation Error in feature {ft_name}: "
                            + f"Feature type {ft_type} is not met by all elements in {data_col}."
                        ),
                    )


def _validate_snowpark_data(data: snowflake.snowpark.DataFrame, features: Sequence[core.BaseFeatureSpec]) -> None:
    """Validate Snowpark DataFrame as input

    Args:
        data: A snowpark dataframe to be validated.
        features: A sequence of feature specifications and feature group specifications, where the dataframe should fit.

    Raises:
        SnowflakeMLException: NotImplementedError: FeatureGroupSpec is not supported.
        SnowflakeMLException: ValueError: Raised when confronting invalid feature.
        SnowflakeMLException: ValueError: Raised when a feature cannot be found.
    """

    schema = data.schema
    for feature in features:
        ft_name = feature.name
        found = False
        for field in schema.fields:
            name = identifier.get_unescaped_names(field.name)
            if name == ft_name:
                found = True
                if field.nullable:
                    warnings.warn(
                        f"Warn in feature {ft_name}: Nullable column {field.name} provided,"
                        + " inference might fail if there is null value.",
                        category=RuntimeWarning,
                        stacklevel=1,
                    )
                if isinstance(feature, core.FeatureGroupSpec):
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.NOT_IMPLEMENTED,
                        original_exception=NotImplementedError("FeatureGroupSpec is not supported."),
                    )
                assert isinstance(feature, core.FeatureSpec)  # mypy
                ft_type = feature._dtype
                field_data_type = field.datatype
                if isinstance(field_data_type, spt.ArrayType):
                    if feature._shape is None:
                        raise snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.INVALID_DATA,
                            original_exception=ValueError(
                                f"Data Validation Error in feature {ft_name}: "
                                + f"Feature is a array feature, while {field.name} is not."
                            ),
                        )
                    warnings.warn(
                        f"Warn in feature {ft_name}: Feature is a array feature, type validation cannot happen.",
                        category=RuntimeWarning,
                        stacklevel=1,
                    )
                else:
                    if feature._shape:
                        raise snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.INVALID_DATA,
                            original_exception=ValueError(
                                f"Data Validation Error in feature {ft_name}: "
                                + f"Feature is a scalar feature, while {field.name} is not."
                            ),
                        )
                    _validate_snowpark_type_feature(data, field, ft_type, ft_name)
        if not found:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(f"Data Validation Error: feature {ft_name} does not exist in data."),
            )


def _validate_snowpark_type_feature(
    df: snowflake.snowpark.DataFrame, field: spt.StructField, ft_type: DataType, ft_name: str
) -> None:
    def get_value_range(field_name: str) -> Tuple[int, int]:
        res = df.select(F.min(field_name).as_("MIN"), F.max(field_name).as_("MAX")).collect()
        if len(res) != 1:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INTERNAL_SNOWML_ERROR,
                original_exception=ValueError(f"Unable to get the value range of field {field_name}"),
            )
        return res[0].MIN, res[0].MAX

    field_data_type = field.datatype
    col_name = identifier.get_unescaped_names(field.name)

    if ft_type in [
        core.DataType.INT8,
        core.DataType.INT16,
        core.DataType.INT32,
        core.DataType.INT64,
        core.DataType.UINT8,
        core.DataType.UINT16,
        core.DataType.UINT32,
        core.DataType.UINT64,
    ]:
        if not (
            isinstance(field_data_type, spt._IntegralType)
            or (isinstance(field_data_type, spt.DecimalType) and field_data_type.scale == 0)
        ):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    f"Data Validation Error in feature {ft_name}: "
                    + f"Feature type {ft_type} is not met by column {col_name}."
                ),
            )
        min_v, max_v = get_value_range(field.name)
        if max_v > np.iinfo(ft_type._numpy_type).max or min_v < np.iinfo(ft_type._numpy_type).min:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    f"Data Validation Error in feature {ft_name}: "
                    + f"Feature type {ft_type} is not met by column {col_name}."
                ),
            )
    elif ft_type in [core.DataType.FLOAT, core.DataType.DOUBLE]:
        if not (
            isinstance(
                field_data_type,
                (spt._IntegralType, spt.FloatType, spt.DoubleType),
            )
            # We are not allowing > 0 scale as it will become a decimal.Decimal
            # Although it is castable, the support will be done as another effort.
            or (isinstance(field_data_type, spt.DecimalType) and field_data_type.scale == 0)
        ):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    f"Data Validation Error in feature {ft_name}: "
                    + f"Feature type {ft_type} is not met by column {col_name}."
                ),
            )
        min_v, max_v = get_value_range(field.name)
        if (
            max_v > np.finfo(ft_type._numpy_type).max  # type: ignore[arg-type]
            or min_v < np.finfo(ft_type._numpy_type).min  # type: ignore[arg-type]
        ):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    f"Data Validation Error in feature {ft_name}: "
                    + f"Feature type {ft_type} is not met by column {col_name}."
                ),
            )
    else:
        if not (isinstance(field_data_type, ft_type._snowpark_type)):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    f"Data Validation Error in feature {ft_name}: "
                    + f"Feature type {ft_type} is not met by column {col_name}."
                ),
            )


def _convert_local_data_to_df(data: model_types.SupportedLocalDataType) -> pd.DataFrame:
    """Convert local data to pandas DataFrame or Snowpark DataFrame

    Args:
        data: The provided data.

    Raises:
        SnowflakeMLException: NotImplementedError: Raised when data cannot be handled by any data handler.

    Returns:
        The converted dataframe with renamed column index.
    """
    df = None
    for handler in _LOCAL_DATA_HANDLERS:
        if handler.can_handle(data):
            handler.validate(data)
            df = handler.convert_to_df(data, ensure_serializable=False)
            break
    if df is None:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.NOT_IMPLEMENTED,
            original_exception=NotImplementedError(f"Data Validation Error: Un-supported type {type(data)} provided."),
        )

    return df


def _convert_and_validate_local_data(
    data: model_types.SupportedLocalDataType, features: Sequence[core.BaseFeatureSpec]
) -> pd.DataFrame:
    """Validate the data with features in model signature and convert to DataFrame

    Args:
        features: A list of feature specs that the data should follow.
        data: The provided data.

    Returns:
        The converted dataframe with renamed column index.
    """
    df = _convert_local_data_to_df(data)
    df = utils.rename_pandas_df(df, features)
    _validate_pandas_df(df, features)
    df = pandas_handler.PandasDataFrameHandler.convert_to_df(df, ensure_serializable=True)

    return df


def infer_signature(
    input_data: model_types.SupportedLocalDataType,
    output_data: model_types.SupportedLocalDataType,
    input_feature_names: Optional[List[str]] = None,
    output_feature_names: Optional[List[str]] = None,
) -> core.ModelSignature:
    """Infer model signature from given input and output sample data.

    Currently, we support infer the model signature from example input/output data in the following cases:
        - Pandas data frame whose column could have types of supported data types,
            list (including list of supported data types, list of numpy array of supported data types, and nested list),
            and numpy array of supported data types.
            - Does not support DataFrame with CategoricalIndex column index.
            - Does not support DataFrame with column of variant length list or numpy array.
        - Numpy array of supported data types.
        - List of Numpy array of supported data types.
        - List of supported data types, or nested list of supported data types.
            - Does not support list of list of variant length list.

    When a ValueError is raised when inferring the signature, it indicates that the data is ill and it is impossible to
    create a signature reflecting that.
    When a NotImplementedError is raised, it indicates that it might be possible to create a signature reflecting the
    provided data, however, we could not infer it.

    Args:
        input_data: Sample input data for the model.
        output_data: Sample output data for the model.
        input_feature_names: Name for input features. Defaults to None.
        output_feature_names: Name for output features. Defaults to None.

    Returns:
        A model signature.
    """
    inputs = _infer_signature(input_data, role="input")
    inputs = utils.rename_features(inputs, input_feature_names)
    outputs = _infer_signature(output_data, role="output")
    outputs = utils.rename_features(outputs, output_feature_names)
    return core.ModelSignature(inputs, outputs)



import inspect
from typing import Any, Callable, Dict, Set, Tuple

import numpy as np
from typing_extensions import TypeGuard

from snowflake.ml._internal.exceptions import error_codes, exceptions
from snowflake.ml.modeling.framework._utils import to_native_format
from snowflake.ml.modeling.framework.base import BaseTransformer
from snowflake.snowpark import Session


def validate_sklearn_args(args: Dict[str, Tuple[Any, Any, bool]], klass: type) -> Dict[str, Any]:
    """Validate if all the keyword args are supported by current version of SKLearn/XGBoost object.

    Args:
        args: Dictionary with kwarg as key. Values is a list with three entries: the kwarg value, default value, and
              whether default is included in signature.
        klass: Underlying SKLearn/XGBoost class object.

    Returns:
        result: sklearn arguments

    Raises:
        SnowflakeMLException: if a user specified arg is not supported by current version of sklearn/xgboost.
    """
    result = {}
    signature = inspect.signature(klass.__init__)  # type: ignore[misc]
    for k, v in args.items():
        if k not in signature.parameters.keys():  # Arg is not supported.
            if v[2] or (  # Arg doesn't have default value in the signature.
                v[0] != v[1]  # Value is not same as default.
                and not (isinstance(v[0], float) and np.isnan(v[0]) and np.isnan(v[1]))
            ):  # both are not NANs
                raise exceptions.SnowflakeMLException(
                    error_code=error_codes.DEPENDENCY_VERSION_ERROR,
                    original_exception=RuntimeError(f"Arg {k} is not supported by current version of SKLearn/XGBoost."),
                )
        else:
            result[k] = v[0]
    return result


def transform_snowml_obj_to_sklearn_obj(obj: Any) -> Any:
    """Converts SnowML Estimator and Transformer objects to equivalent SKLearn objects.

    Args:
        obj: Source object that needs to be converted. Source object could of any type, example, lists, tuples, etc.

    Returns:
        An equivalent object with SnowML estimators and transforms replaced with equivalent SKLearn objects.
    """

    if isinstance(obj, list):
        # Apply transform function to each element in the list
        return list(map(transform_snowml_obj_to_sklearn_obj, obj))
    elif isinstance(obj, tuple):
        # Apply transform function to each element in the tuple
        return tuple(map(transform_snowml_obj_to_sklearn_obj, obj))
    elif isinstance(obj, BaseTransformer):
        # Convert SnowML object to equivalent SKLearn object
        return to_native_format(obj)
    else:
        # Return all other objects as it is.
        return obj


def gather_dependencies(obj: Any) -> Set[str]:
    """Gathers dependencies from the SnowML Estimator and Transformer objects.

    Args:
        obj: Source object to collect dependencies from. Source object could of any type, example, lists, tuples, etc.

    Returns:
        A set of dependencies required to work with the object.
    """

    if isinstance(obj, list) or isinstance(obj, tuple):
        deps: Set[str] = set()
        for elem in obj:
            deps = deps | set(gather_dependencies(elem))
        return deps
    elif isinstance(obj, BaseTransformer):
        return set(obj._get_dependencies())
    else:
        return set()


def original_estimator_has_callable(attr: str) -> Callable[[Any], bool]:
    """Checks that the original estimator has callable `attr`.

    Args:
        attr: Attribute to check for.

    Returns:
        A function which checks for the existence of callable `attr` on the given object.
    """

    def check(self: BaseTransformer) -> TypeGuard[Callable[..., object]]:
        """Check for the existence of callable `attr` in self.

        Args:
            self: BaseTransformer object

        Returns:
            True of the callable `attr` exists in self, False otherwise.
        """
        return callable(getattr(self._sklearn_object, attr, None))

    return check


def is_single_node(session: Session) -> bool:
    """Retrieve the current session's warehouse type and warehouse size, and depends on those information
    to identify if it is single node or not

    Args:
        session (Session): session object that is used by user currently

    Returns:
        bool: single node or not. True stands for yes.
    """
    warehouse_name = session.get_current_warehouse()
    if warehouse_name:
        warehouse_name = warehouse_name.replace('"', "")
        df = session.sql(f"SHOW WAREHOUSES like '{warehouse_name}';")['"type"', '"size"'].collect()[0]
        # filter out the conditions when it is single node
        single_node: bool = (df[0] == "SNOWPARK-OPTIMIZED" and df[1] == "Medium") or (
            df[0] == "STANDARD" and df[1] == "X-Small"
        )
        return single_node
    # If current session cannot retrieve the warehouse name back,
    # Default as True; Let HPO fall back to stored procedure implementation
    return True



import functools
import inspect
from typing import Any, Callable, List, Optional

from snowflake import snowpark
from snowflake.connector import connection
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
    fileset_error_messages,
    fileset_errors,
)
from snowflake.ml._internal.utils import (
    identifier,
    import_utils,
    snowpark_dataframe_utils,
)
from snowflake.ml.fileset import sfcfs
from snowflake.snowpark import exceptions as snowpark_exceptions, functions

# The max file size for data loading.
TARGET_FILE_SIZE = 32 * 2**20

# Expected type of a stage where a FileSet can be located.
# The type is the value of the 'type' column of a `show stages` query.
_FILESET_STAGE_TYPE = "INTERNAL NO CSE"

_PROJECT = "FileSet"


def _raise_if_deleted(func: Callable[..., Any]) -> Callable[..., Any]:
    """A function decorator where an error will be raised when the fileset has been deleted."""

    @functools.wraps(func)
    def raise_if_deleted_helper(self: Any, *args: Any, **kwargs: Any) -> Any:
        if self._is_deleted:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.SNOWML_DELETE_FAILED,
                original_exception=fileset_errors.FileSetAlreadyDeletedError("The FileSet has already been deleted."),
            )
        return func(self, *args, **kwargs)

    return raise_if_deleted_helper


class FileSet:
    """A FileSet represents an immutable snapshot of the result of a query in the form of files."""

    @telemetry.send_api_usage_telemetry(project=_PROJECT)
    def __init__(
        self,
        *,
        target_stage_loc: str,
        name: str,
        sf_connection: Optional[connection.SnowflakeConnection] = None,
        snowpark_session: Optional[snowpark.Session] = None,
    ) -> None:
        """Create a FileSet based on an existing stage directory.

        It can be used to restore an existing FileSet that was not deleted before.

        Args:
            sf_connection: A Snowflake python connection object. Mutually exclusive to `snowpark_session`.
            snowpark_session: A Snowpark Session object. Mutually exclusive to `sf_connection`.
            target_stage_loc: A string of the Snowflake stage path where the FileSet will be stored.
                It needs to be an absolute path with the form of "@{database}.{schema}.{stage}/{optional directory}/".
            name: The name of the FileSet. It is the name of the directory which holds result stage files.

        Raises:
            SnowflakeMLException: An error occurred when not exactly one of sf_connection and snowpark_session is given.

        Example:
        >>> # Create a new FileSet using Snowflake Python connection
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> my_fileset = snowflake.ml.fileset.FileSet.make(
        >>>     target_stage_loc="@mydb.mychema.mystage/mydir"
        >>>     name="helloworld",
        >>>     sf_connection=conn,
        >>>     query="SELECT * FROM Mytable limit 1000000",
        >>> )
        >>> my_fileset.files()
        ----
        ['sfc://@mydb.myschema.mystage//mydir/helloworld/data_0_0_0.snappy.parquet']

        >>> # Now we can restore the FileSet in another program as long as the FileSet is not deleted
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> my_fileset_pointer = FileSet(sf_connection=conn,
                                         target_stage_loc="@mydb.mychema.mystage/mydir",
                                         name="helloworld")
        >>> my_fileset.files()
        ----
        ['sfc://@mydb.myschema.mystage/mydir/helloworld/data_0_0_0.snappy.parquet']
        """
        if sf_connection and snowpark_session:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(fileset_error_messages.BOTH_SF_CONNECTION_AND_SNOWPARK_SESSION_SPECIFIED),
            )
        if not sf_connection and not snowpark_session:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(fileset_error_messages.NO_SF_CONNECTION_OR_SNOWPARK_SESSION),
            )
        self._snowpark_session = (
            snowpark_session
            if snowpark_session
            else snowpark.Session.builder.config("connection", sf_connection).create()
        )
        self._target_stage_loc = target_stage_loc
        _validate_target_stage_loc(self._snowpark_session, self._target_stage_loc)
        self._name = name
        # We want the whole file to be downloaded into memory upon the first head.
        # Because the actual file size might be larger TARGET_FILE_SIZE, we use a larger buffer ceiling.
        self._fs = sfcfs.SFFileSystem(
            snowpark_session=self._snowpark_session,
            cache_type="bytes",
            block_size=2 * TARGET_FILE_SIZE,
        )
        self._files: List[str] = []
        self._is_deleted = False

        _get_fileset_query_id_or_raise(self.files(), self._fileset_absolute_path())

    @classmethod
    def make(
        cls,
        *,
        target_stage_loc: str,
        name: str,
        snowpark_dataframe: Optional[snowpark.DataFrame] = None,
        sf_connection: Optional[connection.SnowflakeConnection] = None,
        query: str = "",
        shuffle: bool = False,
    ) -> "FileSet":
        """Creates a FileSet object given a SQL query.

        The result FileSet object captures the query result deterministically as stage files.

        Args:
            target_stage_loc: A string of the Snowflake stage path where the FileSet will be stored.
                It needs to be an absolute path with the form of "@{database}.{schema}.{stage}/{optional directory}/".
            name: The name of the FileSet. It will become the name of the directory which holds result stage files.
                If there is already a FileSet with the same name in the given stage location,
                an exception will be raised.
            snowpark_dataframe: A Snowpark Dataframe. Mutually exclusive to (`sf_connection`, `query`).
            sf_connection: A Snowflake python connection object. Must be provided if `query` is provided.
            query: A string of Snowflake SQL query to be executed. Mutually exclusive to `snowpark_dataframe`. Must
                also specify `sf_connection`.
            shuffle: A boolean represents whether the data should be shuffled globally. Default to be false.

        Returns:
            A FileSet object.

        Raises:
            ValueError: An error occurred when not exactly one of sf_connection and snowpark_session is given.
            FileSetExistError: An error occurred whern a FileSet with the same name exists in the given path.
            FileSetError: An error occurred when the SQL query/dataframe is not able to get materialized.

        # noqa: DAR401

        Note: During the generation of stage files, data casting will occur. The casting rules are as follows::
            - Data casting:
                - DecimalType(NUMBER):
                    - If its scale is zero, cast to BIGINT
                    - If its scale is non-zero, cast to FLOAT
                - DoubleType(DOUBLE): Cast to FLOAT.
                - ByteType(TINYINT): Cast to SMALLINT.
                - ShortType(SMALLINT):Cast to SMALLINT.
                - IntegerType(INT): Cast to INT.
                - LongType(BIGINT): Cast to BIGINT.
            - No action:
                - FloatType(FLOAT): No action.
                - StringType(String): No action.
                - BinaryType(BINARY): No action.
                - BooleanType(BOOLEAN): No action.
            - Not supported:
                - ArrayType(ARRAY): Not supported. A warning will be logged.
                - MapType(OBJECT): Not supported. A warning will be logged.
                - TimestampType(TIMESTAMP): Not supported. A warning will be logged.
                - TimeType(TIME): Not supported. A warning will be logged.
                - DateType(DATE): Not supported. A warning will be logged.
                - VariantType(VARIANT): Not supported. A warning will be logged.

        Example 1: Create a FileSet with Snowflake Python connection
        --------
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> my_fileset = snowflake.ml.fileset.FileSet.make(
        >>>     target_stage_loc="@mydb.mychema.mystage/mydir"
        >>>     name="helloworld",
        >>>     sf_connection=conn,
        >>>     query="SELECT * FROM mytable limit 1000000",
        >>> )
        >>> my_fileset.files()
        ----
        ['sfc://@mydb.myschema.mystage/helloworld/data_0_0_0.snappy.parquet']

        Example 2: Create a FileSet with a Snowpark dataframe
        --------
        >>> new_session = snowflake.snowpark.Session.builder.configs(connection_parameters).create()
        >>> df = new_session.sql("SELECT * FROM Mytable limit 1000000")
        >>> my_fileset = snowflake.ml.fileset.FileSet.make(
        >>>     target_stage_loc="@mydb.mychema.mystage/mydir"
        >>>     name="helloworld",
        >>>     snowpark_dataframe=df,
        >>> )
        >>> my_fileset.files()
        ----
        ['sfc://@mydb.myschema.mystage/helloworld/data_0_0_0.snappy.parquet']
        """
        if snowpark_dataframe and sf_connection:
            raise ValueError(fileset_error_messages.BOTH_SF_CONNECTION_AND_SNOWPARK_SESSION_SPECIFIED)

        if not snowpark_dataframe:
            if not sf_connection:
                raise ValueError(fileset_error_messages.NO_SF_CONNECTION_OR_SNOWPARK_DATAFRAME)
            if not query:
                raise ValueError("Please use non-empty query to generate meaningful result.")
            snowpark_session = snowpark.Session.builder.config("connection", sf_connection).create()
            snowpark_dataframe = snowpark_session.sql(query)

        assert snowpark_dataframe is not None
        assert snowpark_dataframe._session is not None
        snowpark_session = snowpark_dataframe._session
        casted_df = snowpark_dataframe_utils.cast_snowpark_dataframe(snowpark_dataframe)

        try:
            _validate_target_stage_loc(snowpark_session, target_stage_loc)
        except snowml_exceptions.SnowflakeMLException as e:
            raise e.original_exception
        target_stage_exists = snowpark_session.sql(f"List {_fileset_absolute_path(target_stage_loc, name)}").collect()
        if target_stage_exists:
            raise fileset_errors.FileSetExistError(fileset_error_messages.FILESET_ALREADY_EXISTS.format(name))

        if shuffle:
            casted_df = casted_df.order_by(functions.random())

        try:
            # partition_by helps generate more uniform sharding among files.
            # As a side effect, the sizes of generate files might exceed max_file_size.
            # "partition_by=name" assigns the same sharding key <name> to all rows, resulting in all the generated files
            # located in <target_stage_loc>/<name>/ directory.
            # typing: snowpark's function signature is bogus.
            casted_df.write.copy_into_location(  # type:ignore[call-overload]
                location=target_stage_loc,
                file_format_type="parquet",
                header=True,
                partition_by=f"'{name}'",
                max_file_size=TARGET_FILE_SIZE,
                detailed_output=True,
                statement_params=telemetry.get_function_usage_statement_params(
                    project=_PROJECT,
                    function_name=telemetry.get_statement_params_full_func_name(
                        inspect.currentframe(), cls.__class__.__name__
                    ),
                    api_calls=[snowpark.DataFrameWriter.copy_into_location],
                ),
            )
        except snowpark_exceptions.SnowparkClientException as e:
            # Snowpark wraps the Python Connector error code in the head of the error message.
            if e.message.startswith(fileset_errors.ERRNO_FILE_EXIST_IN_STAGE):
                raise fileset_errors.FileSetExistError(fileset_error_messages.FILESET_ALREADY_EXISTS.format(name))
            else:
                raise fileset_errors.FileSetError(str(e))

        return cls(target_stage_loc=target_stage_loc, name=name, snowpark_session=snowpark_session)

    @property
    def name(self) -> str:
        """Get the name of the FileSet."""
        return self._name

    def _list_files(self) -> List[str]:
        """Private helper function that lists all files in this fileset and caches the results for subsequent use."""
        if self._files:
            return self._files
        loc = self._fileset_absolute_path()

        # TODO(zzhu)[SNOW-703491]: We could use manifest file to speed up file listing
        files = self._fs.ls(loc)
        self._files = [f"sfc://{file}" for file in files]
        return self._files

    def _fileset_absolute_path(self) -> str:
        """Get the Snowflake absolute path to this FileSet directory."""
        return _fileset_absolute_path(self._target_stage_loc, self.name)

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def files(self) -> List[str]:
        """Get the list of stage file paths in the current FileSet.

        The stage file paths follows the sfc protocol.

        Returns:
            A list of stage file paths

        Example:
        >>> my_fileset = FileSet(sf_connection=conn, target_stage_loc="@mydb.mychema.mystage", name="test")
        >>> my_fileset.files()
        ----
        ["sfc://@mydb.myschema.mystage/test/hello_world_0_0_0.snappy.parquet",
         "sfc://@mydb.myschema.mystage/test/hello_world_0_0_1.snappy.parquet"]
        """
        return self._list_files()

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def fileset_stage_location(self) -> str:
        """Get the stage path to the current FileSet in sfc protocol.

        Returns:
            A string representing the stage path

        Example:
        >>> my_fileset = FileSet(sf_connection=conn, target_stage_loc="@mydb.mychema.mystage", name="test")
        >>> my_fileset.files()
        ----
        "sfc://@mydb.myschema.mystage/test/
        """
        location = self._fileset_absolute_path()
        location = "sfc://" + location
        return location

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
        func_params_to_log=["batch_size", "shuffle", "drop_last_batch"],
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def to_torch_datapipe(self, *, batch_size: int, shuffle: bool = False, drop_last_batch: bool = True) -> Any:
        """Transform the Snowflake data into a ready-to-use Pytorch datapipe.

        Return a Pytorch datapipe which iterates on rows of data.

        Args:
            batch_size: It specifies the size of each data batch which will be
                yield in the result datapipe
            shuffle: It specifies whether the data will be shuffled. If True, files will be shuffled, and
                rows in each file will also be shuffled.
            drop_last_batch: Whether the last batch of data should be dropped. If set to be true,
                then the last batch will get dropped if its size is smaller than the given batch_size.

        Returns:
            A Pytorch iterable datapipe that yield data.

        Examples:
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> fileset = FileSet.make(
        >>>     sf_connection=conn, name="helloworld", target_stage_loc="@mydb.myschema.mystage"
        >>>     query="SELECT * FROM Mytable"
        >>> )
        >>> dp = fileset.to_torch_datapipe(batch_size=1)
        >>> for data in dp:
        >>>     print(data)
        ----
        {'_COL_1':[10]}
        """
        IterableWrapper, _ = import_utils.import_or_get_dummy("torchdata.datapipes.iter.IterableWrapper")
        torch_datapipe_module, _ = import_utils.import_or_get_dummy("snowflake.ml.fileset.torch_datapipe")

        self._fs.optimize_read(self._list_files())

        input_dp = IterableWrapper(self._list_files())
        return torch_datapipe_module.ReadAndParseParquet(input_dp, self._fs, batch_size, shuffle, drop_last_batch)

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
        func_params_to_log=["batch_size", "shuffle", "drop_last_batch"],
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def to_tf_dataset(self, *, batch_size: int, shuffle: bool = False, drop_last_batch: bool = True) -> Any:
        """Transform the Snowflake data into a ready-to-use TensorFlow tf.data.Dataset.

        Args:
            batch_size: It specifies the size of each data batch which will be
                yield in the result datapipe
            shuffle: It specifies whether the data will be shuffled. If True, files will be shuffled, and
                rows in each file will also be shuffled.
            drop_last_batch: Whether the last batch of data should be dropped. If set to be true,
                then the last batch will get dropped if its size is smaller than the given batch_size.

        Returns:
            A tf.data.Dataset that yields batched tf.Tensors.

        Examples:
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> fileset = FileSet.make(
        >>>     sf_connection=conn, name="helloworld", target_stage_loc="@mydb.myschema.mystage"
        >>>     query="SELECT * FROM Mytable"
        >>> )
        >>> dp = fileset.to_tf_dataset(batch_size=1)
        >>> for data in dp:
        >>>     print(data)
        ----
        {'_COL_1': <tf.Tensor: shape=(1,), dtype=int64, numpy=[10]>}
        """
        tf_dataset_module, _ = import_utils.import_or_get_dummy("snowflake.ml.fileset.tf_dataset")

        self._fs.optimize_read(self._list_files())

        return tf_dataset_module.read_and_parse_parquet(
            self._list_files(), self._fs, batch_size, shuffle, drop_last_batch
        )

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def to_snowpark_dataframe(self) -> snowpark.DataFrame:
        """Convert the fileset to a snowpark dataframe.

        Only parquet files that owned by the FileSet will be read and converted. The parquet files that materialized by
        FileSet have the name pattern "data_<query_id>_<some_sharding_order>.snappy.parquet".

        Returns:
            A Snowpark dataframe that contains the data of this FileSet.

        Note: The dataframe generated by this method might not have the same schema as the original one. Specifically,
            - NUMBER type with scale != 0 will become float.
            - Unsupported types (see comments of :func:`~FileSet.fileset.make`) will not have any guarantee.
                For example, an OBJECT column may be scanned back as a STRING column.
        """
        query_id = _get_fileset_query_id_or_raise(self._list_files(), self._fileset_absolute_path())
        file_path_pattern = f".*data_{query_id}.*[.]parquet"
        df = self._snowpark_session.read.option("pattern", file_path_pattern).parquet(self._fileset_absolute_path())
        assert isinstance(df, snowpark.DataFrame)
        return df

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    @_raise_if_deleted
    def delete(self) -> None:
        """Delete the FileSet directory and all the stage files in it.

        If not called, the FileSet and all its stage files will stay in Snowflake stage.

        Raises:
            SnowflakeMLException: An error occurred when the FileSet cannot get deleted.
        """
        delete_sql = f"remove {self._fileset_absolute_path()}"
        try:
            self._snowpark_session.sql(delete_sql).collect(
                statement_params=telemetry.get_function_usage_statement_params(
                    project=_PROJECT,
                    function_name=telemetry.get_statement_params_full_func_name(
                        inspect.currentframe(), self.__class__.__name__
                    ),
                ),
            )
            self._files = []
            self._is_deleted = True
        except snowpark_exceptions.SnowparkClientException as e:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.SNOWML_DELETE_FAILED,
                original_exception=fileset_errors.FileSetCannotDeleteError(str(e)),
            )
        return


def _get_fileset_query_id_or_raise(files: List[str], fileset_absolute_path: str) -> Optional[str]:
    """Obtain the query ID used to generate the FileSet stage files.

    If the input stage files are not generated by the same query, an error will be raised.

    Args:
        files: A list of stage file paths follows sfc protocol
        fileset_absolute_path: the Snowflake absolute path to this FileSet directory

    Returns:
        The query id of the sql query which is used to generate the stage files.

    Raises:
        SnowflakeMLException: If the input files are not generated by the same query.
    """
    if not files:
        return None

    valid = True
    query_id = None
    common_prefix = f"sfc://{fileset_absolute_path}data_"
    common_prefix_len = len(common_prefix)
    for file in files:
        if len(file) < common_prefix_len:
            valid = False
            break
        truncatred_filename = file[common_prefix_len:]
        if query_id:
            if not truncatred_filename.startswith(query_id):
                valid = False
                break
        else:
            idx = truncatred_filename.find("_")
            query_id = truncatred_filename[:idx]

    if not valid:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.SNOWML_INVALID_QUERY,
            original_exception=fileset_errors.MoreThanOneQuerySourceError(
                "This FileSet contains files generated by the other queries."
            ),
        )
    return query_id


def _validate_target_stage_loc(snowpark_session: snowpark.Session, target_stage_loc: str) -> bool:
    """Validate the input stage location is in the right format and the target stage is an internal SSE stage.

    A valid format for the input stage location should be '@<database>.<schema>.<stage>/<optional_directories>/',
        where '<database>', '<schema>' and '<stage>' are all snowflake identifiers.

    Args:
        snowpark_session: A snowpark session.
        target_stage_loc: Path to the target location. Should be in the form of
            '@<database>.<schema>.<stage>/<optional_directories>/'

    Returns:
        A Boolean value about whether the input target stage location is a valid path in an internal SSE stage.

    Raises:
        SnowflakeMLException: The input stage path does not start with '@'.
        SnowflakeMLException: No valid stages found.
        SnowflakeMLException: An error occurred when the input stage path is invalid.
    """
    if not target_stage_loc.startswith("@"):
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.SNOWML_INVALID_STAGE,
            original_exception=fileset_errors.FileSetLocationError('FileSet location should start with "@".'),
        )
    try:
        db, schema, stage, _ = identifier.parse_schema_level_object_identifier(target_stage_loc[1:])
        if db is None or schema is None:
            raise ValueError("The stage path should be in the form '@<database>.<schema>.<stage>/*'")
        df_stages = snowpark_session.sql(f"Show stages like '{stage}' in SCHEMA {db}.{schema}")
        df_stages = df_stages.filter(functions.col('"type"').like(f"%{_FILESET_STAGE_TYPE}%"))
        valid_stage = df_stages.collect()
        if not valid_stage:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.SNOWML_NOT_FOUND,
                original_exception=fileset_errors.FileSetLocationError(
                    "A FileSet requires its location to be in an existing server-side-encrypted internal stage."
                    "See https://docs.snowflake.com/en/sql-reference/sql/create-stage#internal-stage-parameters-internalstageparams "  # noqa: E501
                    "on how to create such a stage."
                ),
            )
    except ValueError as e:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.SNOWML_INVALID_STAGE,
            original_exception=fileset_errors.FileSetLocationError(str(e)),
        )
    return True


def _fileset_absolute_path(target_stage_loc: str, fileset_name: str) -> str:
    """Get the Snowflake absolute path to a FileSet.

    Args:
        target_stage_loc: A string of the location where the FileSet lives in. It should be in the form of
            '@<database>.<schema>.<stage>/<optional_directories>/', where
            '<database>', '<schema>' and '<stage>' are all snowflake identifiers.
            A trailing '/' will be added to the location if there is no one present in the given string.
        fileset_name: The name of the FileSet.

    Returns:
        The absolute path to a FileSet in Snowflake. It is supposed to be in the form of
            '@<database>.<schema>.<stage>/<optional_directories>/<fileset_name>/'.
    """
    target_fileset_loc = target_stage_loc
    if not target_fileset_loc.endswith("/"):
        target_fileset_loc += "/"
    target_fileset_loc += fileset_name + "/"
    return target_fileset_loc



import configparser
import os
from typing import Dict, Optional, Union

from absl import logging
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization

from snowflake import snowpark

_DEFAULT_CONNECTION_FILE = "~/.snowsql/config"


def _read_token(token_file: str = "") -> str:
    """
    Reads token from environment or file provided.

    First tries to read the token from environment variable
    (`SNOWFLAKE_TOKEN`) followed by the token file.
    Both the options are tried out in SnowServices.

    Args:
        token_file: File from which token needs to be read. Optional.

    Returns:
        the token.
    """
    token = os.getenv("SNOWFLAKE_TOKEN", "")
    if token:
        return token
    if token_file and os.path.exists(token_file):
        with open(token_file) as f:
            token = f.read()
    return token


_ENCRYPTED_PKCS8_PK_HEADER = b"-----BEGIN ENCRYPTED PRIVATE KEY-----"
_UNENCRYPTED_PKCS8_PK_HEADER = b"-----BEGIN PRIVATE KEY-----"


def _load_pem_to_der(private_key_path: str) -> bytes:
    """Given a private key file path (in PEM format), decode key data into DER format."""
    with open(private_key_path, "rb") as f:
        private_key_pem = f.read()
    private_key_passphrase: Optional[str] = os.getenv("SNOWFLAKE_PRIVATE_KEY_PASSPHRASE", None)

    # Only PKCS#8 format key will be accepted. However, openssl
    # transparently handle PKCS#8 and PKCS#1 format (by some fallback
    # logic) and their is no function to distinguish between them. By
    # reading openssl source code, apparently they also relies on header
    # to determine if give bytes is PKCS#8 format or not
    if not private_key_pem.startswith(_ENCRYPTED_PKCS8_PK_HEADER) and not private_key_pem.startswith(
        _UNENCRYPTED_PKCS8_PK_HEADER
    ):
        raise Exception("Private key provided is not in PKCS#8 format. Please use correct format.")

    if private_key_pem.startswith(_ENCRYPTED_PKCS8_PK_HEADER) and private_key_passphrase is None:
        raise Exception(
            "Private key is encrypted but passphrase could not be found. "
            "Please set SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable."
        )

    if private_key_pem.startswith(_UNENCRYPTED_PKCS8_PK_HEADER):
        private_key_passphrase = None

    private_key = serialization.load_pem_private_key(
        private_key_pem,
        str.encode(private_key_passphrase) if private_key_passphrase is not None else private_key_passphrase,
        backends.default_backend(),
    )

    return private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


def _connection_properties_from_env() -> Dict[str, str]:
    """Returns a dict with all possible login related env variables."""
    sf_conn_prop = {
        # Mandatory fields
        "account": os.environ["SNOWFLAKE_ACCOUNT"],
        "database": os.environ["SNOWFLAKE_DATABASE"],
        # With a default value
        "token_file": os.getenv("SNOWFLAKE_TOKEN_FILE", "/snowflake/session/token"),
        "ssl": os.getenv("SNOWFLAKE_SSL", "on"),
        "protocol": os.getenv("SNOWFLAKE_PROTOCOL", "https"),
    }
    # With empty default value
    for key, env_var in {
        "user": "SNOWFLAKE_USER",
        "authenticator": "SNOWFLAKE_AUTHENTICATOR",
        "password": "SNOWFLAKE_PASSWORD",
        "host": "SNOWFLAKE_HOST",
        "port": "SNOWFLAKE_PORT",
        "schema": "SNOWFLAKE_SCHEMA",
        "warehouse": "SNOWFLAKE_WAREHOUSE",
        "private_key_path": "SNOWFLAKE_PRIVATE_KEY_PATH",
    }.items():
        value = os.getenv(env_var, "")
        if value:
            sf_conn_prop[key] = value
    return sf_conn_prop


def _load_from_snowsql_config_file(connection_name: str, login_file: str = "") -> Dict[str, str]:
    """Loads the dictionary from snowsql config file."""
    snowsql_config_file = login_file if login_file else os.path.expanduser(_DEFAULT_CONNECTION_FILE)
    if not os.path.exists(snowsql_config_file):
        logging.error(f"Connection name given but snowsql config file is not found at: {snowsql_config_file}")
        raise Exception("Snowflake SnowSQL config not found.")

    config = configparser.ConfigParser(inline_comment_prefixes="#")

    if connection_name:
        if not connection_name.startswith("connections."):
            connection_name = "connections." + connection_name
    else:
        # See https://docs.snowflake.com/en/user-guide/snowsql-start.html#configuring-default-connection-settings
        connection_name = "connections"

    logging.info(f"Reading {snowsql_config_file} for connection parameters defined as {connection_name}")
    config.read(snowsql_config_file)
    conn_params = dict(config[connection_name])
    # Remap names to appropriate args in Python Connector API
    # Note: "dbname" should become "database"
    conn_params = {k.replace("name", ""): v.strip('"') for k, v in conn_params.items()}
    if "db" in conn_params:
        conn_params["database"] = conn_params["db"]
        del conn_params["db"]
    return conn_params


@snowpark._internal.utils.private_preview(version="0.2.0")
def SnowflakeLoginOptions(connection_name: str = "", login_file: Optional[str] = None) -> Dict[str, Union[str, bytes]]:
    """Returns a dict that can be used directly into snowflake python connector or Snowpark session config.

    NOTE: Token/Auth information is sideloaded in all cases above, if provided in following order:
      1. If SNOWFLAKE_TOKEN is defined in the environment, it will be used.
      2. If SNOWFLAKE_TOKEN_FILE is defined in the environment and file matching the value found, content of the file
         will be used.

    If token is found, username, password will be reset and 'authenticator' will be set to 'oauth'.

    Python Connector:
    >> ctx = snowflake.connector.connect(**(SnowflakeLoginOptions()))

    Snowpark Session:
    >> session = Session.builder.configs(SnowflakeLoginOptions()).create()

    Usage Note:
      Ideally one should have a snowsql config file. Read more here:
      https://docs.snowflake.com/en/user-guide/snowsql-start.html#configuring-default-connection-settings

    Args:
        connection_name: Name of the connection to look for inside the config file. If `connection_name` is NOT given,
            it tries auth from env variables.
        login_file: If provided, this is used as config file instead of default one (_DEFAULT_CONNECTION_FILE).

    Returns:
        A dict with connection parameters.

    Raises:
        Exception: if none of config file and environment variable are present.
    """
    conn_prop: Dict[str, Union[str, bytes]] = {}
    login_file = login_file or os.path.expanduser(_DEFAULT_CONNECTION_FILE)
    # If login file exists, use this exclusively.
    if os.path.exists(login_file):
        conn_prop = {**(_load_from_snowsql_config_file(connection_name, login_file))}
    else:
        # If environment exists for SNOWFLAKE_ACCOUNT, assume everything
        # comes from environment. Mixing it not allowed.
        account = os.getenv("SNOWFLAKE_ACCOUNT", "")
        if account:
            conn_prop = {**_connection_properties_from_env()}
        else:
            raise Exception("Snowflake credential is neither set in env nor a login file was provided.")

    # Token, if specified, is always side-loaded in all cases.
    token = _read_token(str(conn_prop["token_file"]) if "token_file" in conn_prop else "")
    if token:
        conn_prop["token"] = token
        if "authenticator" not in conn_prop or conn_prop["authenticator"]:
            conn_prop["authenticator"] = "oauth"
    elif "private_key_path" in conn_prop and "private_key" not in conn_prop:
        conn_prop["private_key"] = _load_pem_to_der(str(conn_prop["private_key_path"]))

    if "ssl" in conn_prop and conn_prop["ssl"].lower() == "off":
        conn_prop["protocol"] = "http"

    return conn_prop



import copy
import logging
import posixpath
import tempfile
from types import ModuleType
from typing import IO, List, Optional, Tuple, TypedDict, Union

from typing_extensions import Unpack

from snowflake.ml._internal import env_utils, file_utils
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._deploy_client.warehouse import infer_template
from snowflake.ml.model._packager.model_meta import model_meta
from snowflake.snowpark import session as snowpark_session, types as st

logger = logging.getLogger(__name__)


def _deploy_to_warehouse(
    session: snowpark_session.Session,
    *,
    model_stage_file_path: str,
    model_meta: model_meta.ModelMetadata,
    udf_name: str,
    target_method: str,
    **kwargs: Unpack[model_types.WarehouseDeployOptions],
) -> None:
    """Deploy the model to warehouse as UDF.

    Args:
        session: Snowpark session.
        model_stage_file_path: Path to the stored model zip file in the stage.
        model_meta: Model Metadata.
        udf_name: Name of the UDF.
        target_method: The name of the target method to be deployed.
        **kwargs: Options that control some features in generated udf code.

    Raises:
        SnowflakeMLException: Raised when model file name is unable to encoded using ASCII.
        SnowflakeMLException: Raised when incompatible model.
        SnowflakeMLException: Raised when target method does not exist in model.
        SnowflakeMLException: Raised when confronting invalid stage location.

    """
    # TODO(SNOW-862576): Should remove check on ASCII encoding after SNOW-862576 fixed.
    model_stage_file_name = posixpath.basename(model_stage_file_path)
    if not file_utils._able_ascii_encode(model_stage_file_name):
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_ARGUMENT,
            original_exception=ValueError(
                f"Model file name {model_stage_file_name} cannot be encoded using ASCII. Please rename."
            ),
        )

    relax_version = kwargs.get("relax_version", False)

    if target_method not in model_meta.signatures.keys():
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_ARGUMENT,
            original_exception=ValueError(f"Target method {target_method} does not exist in model."),
        )

    final_packages = _get_model_final_packages(model_meta, session, relax_version=relax_version)

    stage_location = kwargs.get("permanent_udf_stage_location", None)
    if stage_location:
        stage_location = posixpath.normpath(stage_location.strip())
        if not stage_location.startswith("@"):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(f"Invalid stage location {stage_location}."),
            )

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False, encoding="utf-8") as f:
        _write_UDF_py_file(f.file, model_stage_file_name=model_stage_file_name, target_method=target_method, **kwargs)
        logger.info(f"Generated UDF file is persisted at: {f.name}")

        class _UDFParams(TypedDict):
            file_path: str
            func_name: str
            name: str
            input_types: List[st.DataType]
            return_type: st.DataType
            imports: List[Union[str, Tuple[str, str]]]
            packages: List[Union[str, ModuleType]]

        params = _UDFParams(
            file_path=f.name,
            func_name="infer",
            name=udf_name,
            return_type=st.PandasSeriesType(st.MapType(st.StringType(), st.VariantType())),
            input_types=[st.PandasDataFrameType([st.MapType()])],
            imports=[model_stage_file_path],
            packages=list(final_packages),
        )
        if stage_location is None:  # Temporary UDF
            session.udf.register_from_file(**params, replace=True)
        else:  # Permanent UDF
            session.udf.register_from_file(
                **params,
                replace=kwargs.get("replace_udf", False),
                is_permanent=True,
                stage_location=stage_location,
            )

    logger.info(f"{udf_name} is deployed to warehouse.")


def _write_UDF_py_file(
    f: IO[str],
    model_stage_file_name: str,
    target_method: str,
    **kwargs: Unpack[model_types.WarehouseDeployOptions],
) -> None:
    """Generate and write UDF python code into a file

    Args:
        f: File descriptor to write the python code.
        model_stage_file_name: Model zip file name.
        target_method: The name of the target method to be deployed.
        **kwargs: Options that control some features in generated udf code.
    """
    udf_code = infer_template._UDF_CODE_TEMPLATE.format(
        model_stage_file_name=model_stage_file_name,
        _KEEP_ORDER_COL_NAME=infer_template._KEEP_ORDER_COL_NAME,
        target_method=target_method,
        code_dir_name=model_meta.MODEL_CODE_DIR,
    )
    f.write(udf_code)
    f.flush()


def _get_model_final_packages(
    meta: model_meta.ModelMetadata,
    session: snowpark_session.Session,
    relax_version: Optional[bool] = False,
) -> List[str]:
    """Generate final packages list of dependency of a model to be deployed to warehouse.

    Args:
        meta: Model metadata to get dependency information.
        session: Snowpark connection session.
        relax_version: Whether or not relax the version restriction when fail to resolve dependencies.
            Defaults to False.

    Raises:
        SnowflakeMLException: Raised when PIP requirements and dependencies from non-Snowflake anaconda channel found.
        SnowflakeMLException: Raised when not all packages are available in snowflake conda channel.

    Returns:
        List of final packages string that is accepted by Snowpark register UDF call.
    """
    final_packages = None
    if (
        any(channel.lower() not in [env_utils.DEFAULT_CHANNEL_NAME] for channel in meta.env._conda_dependencies.keys())
        or meta.env.pip_requirements
    ):
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.DEPENDENCY_VERSION_ERROR,
            original_exception=RuntimeError(
                "PIP requirements and dependencies from non-Snowflake anaconda channel is not supported."
            ),
        )

    if relax_version:
        relaxed_env = copy.deepcopy(meta.env)
        relaxed_env.relax_version()
        required_packages = relaxed_env._conda_dependencies[env_utils.DEFAULT_CHANNEL_NAME]
    else:
        required_packages = meta.env._conda_dependencies[env_utils.DEFAULT_CHANNEL_NAME]

    final_packages = env_utils.validate_requirements_in_snowflake_conda_channel(
        session, required_packages, python_version=meta.env.python_version
    )

    if final_packages is None:
        relax_version_info_str = "" if relax_version else "Try to set relax_version as True in the options. "
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.DEPENDENCY_VERSION_ERROR,
            original_exception=RuntimeError(
                "The model's dependencyies are not available in Snowflake Anaconda Channel. "
                + relax_version_info_str
                + "Required packages are:\n"
                + " ".join(map(lambda x: f'"{x}"', required_packages))
                + "\n Required Python version is: "
                + meta.env.python_version
            ),
        )
    return final_packages



import os
import posixpath
import shutil
import string
from abc import ABC
from typing import Optional

import importlib_resources

from snowflake.ml._internal import file_utils
from snowflake.ml._internal.utils import identifier
from snowflake.ml.model._deploy_client import image_builds
from snowflake.ml.model._deploy_client.utils import constants
from snowflake.ml.model._packager.model_meta import model_meta
from snowflake.snowpark import FileOperation, Session


class DockerContext(ABC):
    """
    Constructs the Docker context directory required for image building.
    """

    def __init__(
        self,
        context_dir: str,
        model_meta: model_meta.ModelMetadata,
        session: Optional[Session] = None,
        model_zip_stage_path: Optional[str] = None,
    ) -> None:
        """Initialization

        Args:
            context_dir: Path to context directory.
            model_meta: Model Metadata.
            session: Snowpark session.
            model_zip_stage_path: Path to model zip file on stage.
        """
        self.context_dir = context_dir
        self.model_meta = model_meta
        assert (session is None) == (model_zip_stage_path is None)
        self.session = session
        self.model_zip_stage_path = model_zip_stage_path

    def build(self) -> None:
        """
        Generates and/or moves resources into the Docker context directory.Rename the random model directory name to
        constant "model_dir" instead for better readability.
        """
        self._generate_inference_code()
        self._copy_entrypoint_script_to_docker_context()
        self._copy_model_env_dependency_to_docker_context()
        self._generate_docker_file()

    def _copy_entrypoint_script_to_docker_context(self) -> None:
        """Copy gunicorn_run.sh entrypoint to docker context directory."""
        with importlib_resources.as_file(
            importlib_resources.files(image_builds).joinpath(  # type: ignore[no-untyped-call]
                constants.ENTRYPOINT_SCRIPT
            )
        ) as path:
            shutil.copy(path, os.path.join(self.context_dir, constants.ENTRYPOINT_SCRIPT))

    def _copy_model_env_dependency_to_docker_context(self) -> None:
        """
        Convert model dependencies to files from model metadata.
        """
        self.model_meta.save(self.context_dir)

    def _generate_docker_file(self) -> None:
        """
        Generates dockerfile based on dockerfile template.
        """
        docker_file_path = os.path.join(self.context_dir, "Dockerfile")
        docker_file_template = (
            importlib_resources.files(image_builds)
            .joinpath("templates/dockerfile_template")  # type: ignore[no-untyped-call]
            .read_text("utf-8")
        )

        if self.model_zip_stage_path is not None:
            norm_stage_path = posixpath.normpath(identifier.remove_prefix(self.model_zip_stage_path, "@"))
            assert self.session
            fop = FileOperation(self.session)
            # The explicit download here is inefficient but a compromise.
            # We could in theory reuse the download needed for metadata extraction, but it's hacky and will go away.
            # Ideally, the model download should happen as part of the server side image build,
            # but it requires have our own image builder since there's need to be logic downloading model
            # into the context directory.
            get_res_list = fop.get(stage_location=self.model_zip_stage_path, target_directory=self.context_dir)
            assert len(get_res_list) == 1, f"Single zip file should be returned, but got {len(get_res_list)} files."
            local_zip_file_path = os.path.basename(get_res_list[0].file)
            copy_model_statement = f"COPY {local_zip_file_path} ./{norm_stage_path}"
            extra_env_statement = f"ENV MODEL_ZIP_STAGE_PATH={norm_stage_path}"
        else:
            copy_model_statement = ""
            extra_env_statement = ""

        with open(docker_file_path, "w", encoding="utf-8") as dockerfile:
            base_image = "mambaorg/micromamba:1.4.3"
            tag = base_image.split(":")[1]
            assert tag != constants.LATEST_IMAGE_TAG, (
                "Base image tag should not be 'latest' as it might cause false" "positive image cache hit"
            )
            dockerfile_content = string.Template(docker_file_template).safe_substitute(
                {
                    "base_image": "mambaorg/micromamba:1.4.3",
                    "model_env_folder": constants.MODEL_ENV_FOLDER,
                    "inference_server_dir": constants.INFERENCE_SERVER_DIR,
                    "entrypoint_script": constants.ENTRYPOINT_SCRIPT,
                    # Instead of omitting this ENV var when no CUDA required, we explicitly set it to empty to override
                    # as no CUDA is detected thus it won't be affected by the existence of CUDA in base image.
                    # https://conda.io/projects/conda/en/latest/user-guide/tasks/manage-virtual.html
                    "cuda_override_env": self.model_meta.env.cuda_version if self.model_meta.env.cuda_version else "",
                    "copy_model_statement": copy_model_statement,
                    "extra_env_statement": extra_env_statement,
                }
            )
            dockerfile.write(dockerfile_content)

    def _generate_inference_code(self) -> None:
        """
        Generates inference code based on the app template and creates a folder named 'server' to house the inference
        server code.
        """
        with importlib_resources.as_file(
            importlib_resources.files(image_builds).joinpath(  # type: ignore[no-untyped-call]
                constants.INFERENCE_SERVER_DIR
            )
        ) as inference_server_folder_path:
            destination_folder_path = os.path.join(self.context_dir, constants.INFERENCE_SERVER_DIR)
            ignore_patterns = shutil.ignore_patterns("BUILD.bazel", "*test.py", "*.\\.*", "__pycache__")
            file_utils.copytree(
                inference_server_folder_path,
                destination_folder_path,
                ignore=ignore_patterns,
            )



#!/usr/bin/env python3
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
from typing import Iterable, Optional, Union

import pandas as pd
from sklearn import preprocessing

from snowflake import snowpark
from snowflake.ml._internal import telemetry, type_utils
from snowflake.ml._internal.exceptions import error_codes, exceptions
from snowflake.ml.modeling.framework import base
from snowflake.ml.modeling.preprocessing import ordinal_encoder


class LabelEncoder(base.BaseTransformer):
    r"""Encodes target labels with values between 0 and n_classes-1.

    In other words, each class (i.e., distinct numeric or string) is assigned an integer value, starting with zero.
    LabelEncoder is a specialization of OrdinalEncoder for 1-dimensional data.

    For more details on what this transformer does, see [sklearn.preprocessing.LabelEncoder]
    (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html).

    Args:
        input_cols: The name of a column in a DataFrame to be encoded. May be a string or a list containing one string.
        output_cols: The name of a column in a DataFrame where the results will be stored. May be a string or a list
            containing one string.
        drop_input_cols: Remove input columns from output if set True. False by default.
    """

    def __init__(
        self,
        input_cols: Optional[Union[str, Iterable[str]]] = None,
        output_cols: Optional[Union[str, Iterable[str]]] = None,
        drop_input_cols: Optional[bool] = False,
    ) -> None:
        """
        Encode target labels with integers between 0 and n_classes-1.

        Args:
            input_cols: The name of a column in a DataFrame to be encoded. May be a string or a list containing one
                string.
            output_cols: The name of a column in a DataFrame where the results will be stored. May be a string or a list
                containing one string.
            drop_input_cols: Remove input columns from output if set True. False by default.

        Attributes:
            classes_: A np.ndarray that holds the label for each class.
                Attributes are valid only after fit() has been called.

        """
        super().__init__(drop_input_cols=drop_input_cols)
        self._ordinal_encoder: Optional[ordinal_encoder.OrdinalEncoder] = None
        self.classes_: Optional[type_utils.LiteralNDArrayType] = None
        self.set_input_cols(input_cols)
        self.set_output_cols(output_cols)

    def _reset(self) -> None:
        super()._reset()
        if self._ordinal_encoder:
            self._ordinal_encoder = None
            self.classes_ = None

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def fit(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> "LabelEncoder":
        """
        Fit label encoder with label column in dataset.

        Validates the transformer arguments and derives the list of classes (distinct values) from the data, making
        this list available as an attribute of the transformer instance (see Attributes).
        Returns the transformer instance.

        Args:
            dataset: Input dataset.

        Returns:
            self

        Raises:
            SnowflakeMLException: If length of input_cols is not 1 or length of output_cols is greater than 1.
        """
        if len(self.input_cols) != 1:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError("Label encoder must specify one input column."),
            )
        input_col = self.input_cols[0]

        if len(self.output_cols) != 1:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError("Label encoder must specify one output column."),
            )

        self._reset()

        # Use `OrdinalEncoder` to handle fits and transforms.
        self._ordinal_encoder = ordinal_encoder.OrdinalEncoder(input_cols=self.input_cols, output_cols=self.output_cols)

        self._ordinal_encoder.fit(dataset)

        # Set `classes_` for compatibility with sklearn.
        self.classes_ = self._ordinal_encoder.categories_[input_col]

        self._is_fitted = True
        return self

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    @telemetry.add_stmt_params_to_df(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def transform(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> Union[snowpark.DataFrame, pd.DataFrame]:
        """
        Use fit result to transform snowpark dataframe or pandas dataframe. The original dataset with
        the transform result column added will be returned.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset.
        """
        self._enforce_fit()
        super()._check_dataset_type(dataset)

        if isinstance(dataset, snowpark.DataFrame):
            # [SNOW-802691] Support for mypy type checking
            assert self._ordinal_encoder is not None
            output_df = self._ordinal_encoder.transform(dataset).na.replace(
                float("nan"),
                len(self.classes_) - 1,  # type: ignore[arg-type]
                subset=self.output_cols,
            )
        else:
            output_df = self._transform_sklearn(dataset)

        return self._drop_input_columns(output_df) if self._drop_input_cols is True else output_df

    def _create_unfitted_sklearn_object(self) -> preprocessing.LabelEncoder:
        return preprocessing.LabelEncoder()

    def _create_sklearn_object(self) -> preprocessing.LabelEncoder:
        """
        Initialize and return the equivalent sklearn label encoder.

        Returns:
            Equivalent sklearn object.
        """
        label_encoder = self._create_unfitted_sklearn_object()
        label_encoder.classes_ = self.classes_
        return label_encoder



from typing import Any, Dict

from packaging import requirements, version

from snowflake.ml._internal import env as snowml_env
from snowflake.ml.model._packager.model_meta_migrator import base_migrator


class MetaMigrator_v1(base_migrator.BaseModelMetaMigrator):
    source_version = "1"
    target_version = "2023-12-01"

    @staticmethod
    def upgrade(original_meta_dict: Dict[str, Any]) -> Dict[str, Any]:
        loaded_python_version = version.parse(original_meta_dict["python_version"])
        if original_meta_dict.get("local_ml_library_version", None):
            loaded_lib_version = str(version.parse(original_meta_dict["local_ml_library_version"]))
        else:
            lib_spec_str = next(
                filter(
                    lambda x: requirements.Requirement(x).name == "snowflake-ml-python",
                    original_meta_dict["conda_dependencies"],
                ),
                None,
            )
            if lib_spec_str is None:
                loaded_lib_version = snowml_env.VERSION
            loaded_lib_version = list(requirements.Requirement(str(lib_spec_str)).specifier)[0].version

        return dict(
            creation_timestamp=original_meta_dict["creation_timestamp"],
            env=dict(
                conda="env/conda.yaml",
                pip="env/requirements.txt",
                python_version=f"{loaded_python_version.major}.{loaded_python_version.minor}",
                cuda_version=original_meta_dict.get("cuda_version", None),
                snowpark_ml_version=loaded_lib_version,
            ),
            metadata=original_meta_dict.get("metadata", None),
            model_type=original_meta_dict["model_type"],
            models={
                name: {**value, "handler_version": "2023-12-01"} for name, value in original_meta_dict["models"].items()
            },
            name=original_meta_dict["name"],
            signatures=original_meta_dict["signatures"],
            version=MetaMigrator_v1.target_version,
            min_snowpark_ml_version="1.0.12",
        )



from typing import Optional, Union

from snowflake import snowpark
from snowflake.cortex._util import CORTEX_FUNCTIONS_TELEMETRY_PROJECT, call_sql_function
from snowflake.ml._internal import telemetry


@snowpark._internal.utils.experimental(version="1.0.12")
@telemetry.send_api_usage_telemetry(
    project=CORTEX_FUNCTIONS_TELEMETRY_PROJECT,
)
def Translate(
    text: Union[str, snowpark.Column],
    from_language: Union[str, snowpark.Column],
    to_language: Union[str, snowpark.Column],
    session: Optional[snowpark.Session] = None,
) -> Union[str, snowpark.Column]:
    """Translate calls into the LLM inference service to perform translation.

    Args:
        text: A Column of strings to translate.
        from_language: A Column of input languages.
        to_language: A Column of output languages.
        session: The snowpark session to use. Will be inferred by context if not specified.

    Returns:
        A column of string translations.
    """

    return _translate_impl("snowflake.ml.translate", text, from_language, to_language, session=session)


def _translate_impl(
    function: str,
    text: Union[str, snowpark.Column],
    from_language: Union[str, snowpark.Column],
    to_language: Union[str, snowpark.Column],
    session: Optional[snowpark.Session] = None,
) -> Union[str, snowpark.Column]:
    return call_sql_function(function, session, text, from_language, to_language)



import logging
from typing import Optional

logger = logging.getLogger(__name__)


class LogStreamProcessor:
    def __init__(self) -> None:
        self.last_line_seen = 0

    def process_new_logs(self, job_logs: Optional[str], *, log_level: int = logging.INFO) -> None:
        if not job_logs:
            return
        log_entries = job_logs.split("\n")
        start_index = self.last_line_seen
        log_length = len(log_entries)
        for i in range(start_index, log_length):
            log_entry = log_entries[i]
            if log_level == logging.DEBUG:
                logger.debug(log_entry)
            elif log_level == logging.INFO:
                logger.info(log_entry)
            elif log_level == logging.WARNING:
                logger.warning(log_entry)
            elif log_level == logging.ERROR:
                logger.error(log_entry)
            elif log_level == logging.CRITICAL:
                logger.critical(log_entry)

        self.last_line_seen = log_length



"""
A minimal pure python library to copy images between two remote registries.

This library only supports a limited set of features:
- Works only with docker and OCI manifests and manifest lists for multiarch images (most newer images)
  - Supported OCI manifest type: application/vnd.oci.image.manifest.v1+json
  - Supported Docker manifest type: application/vnd.docker.distribution.manifest.v2+json
- Supports only pulling a single architecture from a multiarch image. Does not support pulling all architectures.
- Supports only schemaVersion 2.
- Streams images from source to destination without any intermediate disk storage in chunks.
- Does not support copying in parallel.

It's recommended to use this library to copy previously tested images using sha256 to avoid surprises
with respect to compatibility.
"""
import dataclasses
import hashlib
import io
import json
import logging
from collections import namedtuple
from typing import Dict, List, Optional, Tuple

import requests

from snowflake.ml._internal.utils import image_registry_http_client

# Common HTTP headers
_CONTENT_LENGTH_HEADER = "content-length"
_CONTENT_TYPE_HEADER = "content-type"
_CONTENT_RANGE_HEADER = "content-range"
_LOCATION_HEADER = "location"
_AUTHORIZATION_HEADER = "Authorization"
_ACCEPT_HEADER = "accept"

_OCI_MANIFEST_LIST_TYPE = "application/vnd.oci.image.index.v1+json"
_DOCKER_MANIFEST_LIST_TYPE = "application/vnd.docker.distribution.manifest.list.v2+json"

_OCI_MANIFEST_TYPE = "application/vnd.oci.image.manifest.v1+json"
_DOCKER_MANIFEST_TYPE = "application/vnd.docker.distribution.manifest.v2+json"

ALL_SUPPORTED_MEDIA_TYPES = [
    _OCI_MANIFEST_LIST_TYPE,
    _DOCKER_MANIFEST_LIST_TYPE,
    _OCI_MANIFEST_TYPE,
    _DOCKER_MANIFEST_TYPE,
]
_MANIFEST_SUPPORTED_KEYS = {"schemaVersion", "mediaType", "config", "layers"}

# Architecture descriptor as a named tuple
_Arch = namedtuple("_Arch", ["arch_name", "os"])

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ImageDescriptor:
    """
    Create an image descriptor.

    registry_name: the name of the registry like gcr.io
    repository_name: the name of the repository like kaniko-project/executor
    tag: the tag of the image like v1.6.0
    digest: the sha256 digest of the image like sha256:b8c0...
    protocol: the protocol to use, defaults to https

    Only a tag or a digest must be specified, not both.
    """

    registry_name: str
    repository_name: str
    tag: Optional[str] = None
    digest: Optional[str] = None
    protocol: str = "https"

    def __baseurl(self) -> str:
        return f"{self.protocol}://{self.registry_name}/v2/"

    def manifest_link(self) -> str:
        return f"{self.__baseurl()}{self.repository_name}/manifests/{self.tag or self.digest}"

    def blob_link(self, digest: str) -> str:
        return f"{self.__baseurl()}{self.repository_name}/blobs/{digest}"

    def blob_upload_link(self) -> str:
        return f"{self.__baseurl()}{self.repository_name}/blobs/uploads/"

    def manifest_upload_link(self, tag: str) -> str:
        return f"{self.__baseurl()}{self.repository_name}/manifests/{tag}"

    def __str__(self) -> str:
        return f"{self.registry_name}/{self.repository_name}@{self.tag or self.digest}"


class Manifest:
    def __init__(self, manifest_bytes: bytes, manifest_digest: str) -> None:
        """Create a manifest object from the manifest JSON dict.

        Args:
            manifest_bytes: manifest content in bytes.
            manifest_digest: SHA256 digest.
        """
        self.manifest_bytes = manifest_bytes
        self.manifest = json.loads(manifest_bytes.decode("utf-8"))
        self.__validate(self.manifest)

        self.manifest_digest = manifest_digest
        self.media_type = self.manifest["mediaType"]

    def get_blob_digests(self) -> List[str]:
        """
        Get the list of blob digests from the manifest including config and layers.
        """
        blobs = []
        blobs.extend([x["digest"] for x in self.manifest["layers"]])
        blobs.append(self.manifest["config"]["digest"])

        return blobs

    def __validate(self, manifest: Dict[str, str]) -> None:
        """
        Validate the manifest.
        """
        assert (
            manifest.keys() == _MANIFEST_SUPPORTED_KEYS
        ), f"Manifest must contain all keys and no more {_MANIFEST_SUPPORTED_KEYS}"
        assert int(manifest["schemaVersion"]) == 2, "Only manifest schemaVersion 2 is supported"
        assert manifest["mediaType"] in [
            _OCI_MANIFEST_TYPE,
            _DOCKER_MANIFEST_TYPE,
        ], f'Unsupported mediaType {manifest["mediaType"]}'

    def __str__(self) -> str:
        """
        Return the manifest as a string.
        """
        return json.dumps(self.manifest, indent=4)


@dataclasses.dataclass
class BlobTransfer:
    """
    Helper class to transfer a blob from one registry to another
    in small chunks using in-memory buffering.
    """

    # Uploads in chunks of 1MB
    chunk_size_bytes = 1024 * 1024

    src_image: ImageDescriptor
    dest_image: ImageDescriptor
    manifest: Manifest
    image_registry_http_client: image_registry_http_client.ImageRegistryHttpClient

    def upload_all_blobs(self) -> None:
        blob_digests = self.manifest.get_blob_digests()
        logger.debug(f"Found {len(blob_digests)} blobs for {self.src_image}")

        for blob_digest in blob_digests:
            logger.debug(f"Transferring blob {blob_digest} from {self.src_image} to {self.dest_image}")
            if self._should_upload(blob_digest):
                self._transfer(blob_digest)
            else:
                logger.debug(f"Blob {blob_digest} already exists in {self.dest_image}")

    def _should_upload(self, blob_digest: str) -> bool:
        """
        Check if the blob already exists in the destination registry.
        """
        resp = self.image_registry_http_client.head(self.dest_image.blob_link(blob_digest), headers={})
        return resp.status_code != 200

    def _fetch_blob(self, blob_digest: str) -> Tuple[io.BytesIO, int]:
        """
        Fetch a stream to the blob from the source registry.
        """
        src_blob_link = self.src_image.blob_link(blob_digest)
        headers = {_CONTENT_LENGTH_HEADER: "0"}
        resp = self.image_registry_http_client.get(src_blob_link, headers=headers)

        assert resp.status_code == 200, f"Blob GET failed with code {resp.status_code}"
        assert _CONTENT_LENGTH_HEADER in resp.headers, f"Blob does not contain {_CONTENT_LENGTH_HEADER}"

        return io.BytesIO(resp.content), int(resp.headers[_CONTENT_LENGTH_HEADER])

    def _get_upload_url(self) -> str:
        """
        Obtain the upload URL from the destination registry.
        """
        response = self.image_registry_http_client.post(self.dest_image.blob_upload_link())
        assert (
            response.status_code == 202
        ), f"Failed to get the upload URL to destination. Status {response.status_code}. {str(response.content)}"
        return str(response.headers[_LOCATION_HEADER])

    def _upload_blob(self, blob_digest: str, blob_data: io.BytesIO, content_length: int) -> None:
        """
        Upload a blob to the destination registry.
        """
        upload_url = self._get_upload_url()
        headers = {
            _CONTENT_TYPE_HEADER: "application/octet-stream",
        }

        # Use chunked transfer
        # This can be optimized to use a single PUT request for small blobs
        next_loc = upload_url
        start_byte = 0
        while start_byte < content_length:
            chunk = blob_data.read(self.chunk_size_bytes)
            chunk_length = len(chunk)
            end_byte = start_byte + chunk_length - 1

            headers[_CONTENT_RANGE_HEADER] = f"{start_byte}-{end_byte}"
            headers[_CONTENT_LENGTH_HEADER] = str(chunk_length)

            resp = self.image_registry_http_client.patch(next_loc, headers=headers, data=chunk)
            assert resp.status_code == 202, f"Blob PATCH failed with code {resp.status_code}"

            next_loc = resp.headers[_LOCATION_HEADER]
            start_byte += chunk_length

        # Finalize the upload
        resp = self.image_registry_http_client.put(f"{next_loc}&digest={blob_digest}")
        assert resp.status_code == 201, f"Blob PUT failed with code {resp.status_code}"

    def _transfer(self, blob_digest: str) -> None:
        """
        Transfer a blob from the source registry to the destination registry.
        """
        blob_data, content_length = self._fetch_blob(blob_digest)
        self._upload_blob(blob_digest, blob_data, content_length)


def get_bytes_with_sha_verification(resp: requests.Response, sha256_digest: str) -> Tuple[bytes, str]:
    """Get the bytes of a response and verify the sha256 digest.

    Args:
        resp: the response object
        sha256_digest: the expected sha256 digest in format "sha256:b8c0..."

    Returns:
        (res, sha256_digest)

    """
    digest = hashlib.sha256()
    chunks = []
    for chunk in resp.iter_content(chunk_size=8192):
        digest.update(chunk)
        chunks.append(chunk)

    calculated_digest = digest.hexdigest()
    assert not sha256_digest or sha256_digest.endswith(calculated_digest), "SHA256 digest does not match"

    content = b"".join(chunks)  # Minimize allocations by joining chunks
    return content, calculated_digest


def get_manifest(
    image_descriptor: ImageDescriptor, arch: _Arch, retryable_http: image_registry_http_client.ImageRegistryHttpClient
) -> Manifest:
    """Get the manifest of an image from the remote registry.

    Args:
        image_descriptor: the image descriptor
        arch: the architecture to filter for if it's a multi-arch image
        retryable_http: a retryable http client.

    Returns:
        Manifest object.

    """
    logger.debug(f"Getting manifest from {image_descriptor.manifest_link()}")

    headers = {_ACCEPT_HEADER: ",".join(ALL_SUPPORTED_MEDIA_TYPES)}

    response = retryable_http.get(image_descriptor.manifest_link(), headers=headers)
    assert response.status_code == 200, f"Manifest GET failed with code {response.status_code}, {response.text}"

    assert image_descriptor.digest
    manifest_bytes, manifest_digest = get_bytes_with_sha_verification(response, image_descriptor.digest)
    manifest_json = json.loads(manifest_bytes.decode("utf-8"))

    # If this is a manifest list, find the manifest for the specified architecture
    # and recurse till we find the real manifest
    if manifest_json["mediaType"] in [
        _OCI_MANIFEST_LIST_TYPE,
        _DOCKER_MANIFEST_LIST_TYPE,
    ]:
        logger.debug("Found a multiarch image. Following manifest reference.")

        assert "manifests" in manifest_json, "Manifest list does not contain manifests"
        qualified_manifests = [
            x
            for x in manifest_json["manifests"]
            if x["platform"]["architecture"] == arch.arch_name and x["platform"]["os"] == arch.os
        ]
        assert (
            len(qualified_manifests) == 1
        ), "Manifest list does not contain exactly one qualified manifest for this arch"

        manifest_object = qualified_manifests[0]
        manifest_digest = manifest_object["digest"]

        logger.debug(f"Found manifest reference for arch {arch}: {manifest_digest}")

        # Copy the image descriptor to fetch the arch-specific manifest
        descriptor_copy = ImageDescriptor(
            registry_name=image_descriptor.registry_name,
            repository_name=image_descriptor.repository_name,
            digest=manifest_digest,
            tag=None,
        )

        # Supports only one level of manifest list nesting to avoid infinite recursion
        return get_manifest(descriptor_copy, arch, retryable_http)

    return Manifest(manifest_bytes, manifest_digest)


def put_manifest(
    image_descriptor: ImageDescriptor,
    manifest: Manifest,
    retryable_http: image_registry_http_client.ImageRegistryHttpClient,
) -> None:
    """
    Upload the given manifest to the destination registry.
    """
    assert image_descriptor.tag is not None, "Tag must be specified for manifest upload"
    headers = {_CONTENT_TYPE_HEADER: manifest.media_type}
    url = image_descriptor.manifest_upload_link(image_descriptor.tag)
    logger.debug(f"Uploading manifest to {url}")
    response = retryable_http.put(url, headers=headers, data=manifest.manifest_bytes)
    assert response.status_code == 201, f"Manifest PUT failed with code {response.status_code}"


def copy_image(
    src_image: ImageDescriptor,
    dest_image: ImageDescriptor,
    arch: _Arch,
    retryable_http: image_registry_http_client.ImageRegistryHttpClient,
) -> None:
    logger.debug(f"Pulling image manifest for {src_image}")

    # 1. Get the manifest
    manifest = get_manifest(src_image, arch, retryable_http)
    logger.debug(f"Manifest pulled for {src_image} with digest {manifest.manifest_digest}")

    # 2: Retrieve all blob digests from manifest; fetch blob based on blob digest, then upload blob.
    blob_transfer = BlobTransfer(src_image, dest_image, manifest, image_registry_http_client=retryable_http)
    blob_transfer.upload_all_blobs()

    # 3. Upload the manifest
    logger.debug(f"All blobs copied successfully. Copying manifest for {src_image} to {dest_image}")
    put_manifest(dest_image, manifest, retryable_http)

    logger.debug(f"Image {src_image} copied to {dest_image}")


def convert_to_image_descriptor(
    image_name: str,
    with_digest: bool = False,
    with_tag: bool = False,
) -> ImageDescriptor:
    """Convert a full image name to a ImageDescriptor object.

    Args:
        image_name: name of image.
        with_digest: boolean to specify whether a digest is included in the image name
        with_tag: boolean to specify whether a tag is included in the image name.

    Returns:
        An ImageDescriptor instance
    """
    assert with_digest or with_tag, "image should contain either digest or tag"
    sep = "@" if with_digest else ":"
    parts = image_name.split("/")
    assert len(parts[-1].split(sep)) == 2, f"Image {image_name} missing digest/tag"
    tag_digest = parts[-1].split(sep)[1]
    return ImageDescriptor(
        registry_name=parts[0],
        repository_name="/".join(parts[1:-1] + [parts[-1].split(sep)[0]]),
        digest=tag_digest if with_digest else None,
        tag=tag_digest if with_tag else None,
    )



import json
import logging
import os
import shutil
import subprocess
import tempfile
import time
from enum import Enum
from typing import List

from snowflake import snowpark
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import spcs_image_registry
from snowflake.ml.model._deploy_client.image_builds import base_image_builder

logger = logging.getLogger(__name__)


class Platform(Enum):
    LINUX_AMD64 = "linux/amd64"


class ClientImageBuilder(base_image_builder.ImageBuilder):
    """
    Client-side image building and upload to model registry.

    Usage requirements:
    Requires prior installation and running of Docker with BuildKit. See installation instructions in
        https://docs.docker.com/engine/install/
    """

    def __init__(
        self,
        *,
        context_dir: str,
        full_image_name: str,
        image_repo: str,
        session: snowpark.Session,
    ) -> None:
        """Initialization

        Args:
            context_dir: Local docker context dir.
            full_image_name: Full image name consists of image name and image tag.
            image_repo: Path to image repository.
            session: Snowpark session
        """
        self.context_dir = context_dir
        self.full_image_name = full_image_name
        self.image_repo = image_repo
        self.session = session

    def build_and_upload_image(self) -> None:
        """Builds and uploads an image to the model registry.

        Raises:
            SnowflakeMLException: Occurs when failed to build image or push to image registry.
        """

        def _setup_docker_config(docker_config_dir: str, registry_cred: str) -> None:
            """Set up a temporary docker config, which is used for running all docker commands. The format of config
            is based on the format that is compatible with docker credential helper:
            {
              "auths": {
                "https://index.docker.io/v1/": {
                  "auth": "<base_64_encoded_username_password>"
                }
              }
            }

            Docker will try to find cli-plugins in the config dir, and then fallback to
            /usr/local/lib/docker/cli-plugins OR /usr/local/libexec/docker/cli-plugins
            /usr/lib/docker/cli-plugins OR /usr/libexec/docker/cli-plugins
            To prevent the case that none of them exists, we copy the cli-plugins in the current config directory,
            which is defined in DOCKER_CONFIG and default to $HOME/.docker to our temp config dir.

            Args:
                docker_config_dir: Path to docker configuration directory, which stores the temporary session token.
                registry_cred: image registry basic auth credential.
            """
            orig_docker_config_dir = os.getenv("DOCKER_CONFIG", os.path.join(os.path.expanduser("~"), ".docker"))
            if os.path.exists(os.path.join(orig_docker_config_dir, "cli-plugins")):
                shutil.copytree(
                    os.path.join(orig_docker_config_dir, "cli-plugins"),
                    os.path.join(docker_config_dir, "cli-plugins"),
                    symlinks=True,
                )
            content = {"auths": {self.full_image_name: {"auth": registry_cred}}}
            config_path = os.path.join(docker_config_dir, "config.json")
            with open(config_path, "w", encoding="utf-8") as file:
                json.dump(content, file)

        def _cleanup_local_image(docker_config_dir: str) -> None:
            try:
                image_exist_command = ["docker", "image", "inspect", self.full_image_name]
                self._run_docker_commands(image_exist_command)
            except Exception:
                # Image does not exist, probably due to failed build step
                pass
            else:
                commands = ["docker", "--config", docker_config_dir, "rmi", self.full_image_name]
                logger.debug(f"Removing local image: {self.full_image_name}")
                self._run_docker_commands(commands)

        self.validate_docker_client_env()
        with spcs_image_registry.generate_image_registry_credential(
            self.session
        ) as registry_cred, tempfile.TemporaryDirectory() as docker_config_dir:
            try:
                _setup_docker_config(docker_config_dir=docker_config_dir, registry_cred=registry_cred)
                start = time.time()
                self._build_and_tag(docker_config_dir)
                end = time.time()
                logger.info(f"Time taken to build the image on the client: {end - start:.2f} seconds")

            except Exception as e:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INTERNAL_DOCKER_ERROR,
                    original_exception=RuntimeError("Failed to build docker image."),
                ) from e
            else:
                try:
                    start = time.time()
                    self._upload(docker_config_dir)
                    end = time.time()
                    logger.info(f"Time taken to upload the image to image registry: {end - start:.2f} seconds")
                except Exception as e:
                    raise snowml_exceptions.SnowflakeMLException(
                        error_code=error_codes.INTERNAL_DOCKER_ERROR,
                        original_exception=RuntimeError("Failed to upload docker image to registry."),
                    ) from e
                finally:
                    _cleanup_local_image(docker_config_dir)

    def validate_docker_client_env(self) -> None:
        """Ensure docker client is running and BuildKit is enabled. Note that Buildx always uses BuildKit.
        - Ensure docker daemon is running through the "docker info" command on shell. When docker daemon is running,
        return code will be 0, else return code will be 1.
        - Ensure BuildKit is enabled by checking "docker buildx version".

        Raises:
            SnowflakeMLException: Occurs when Docker is not installed or is not running.

        """
        try:
            self._run_docker_commands(["docker", "info"])
        except Exception:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.CLIENT_DEPENDENCY_MISSING_ERROR,
                original_exception=ConnectionError(
                    "Failed to initialize Docker client. Please ensure Docker is installed and running."
                ),
            )

        try:
            self._run_docker_commands(["docker", "buildx", "version"])
        except Exception:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.CLIENT_DEPENDENCY_MISSING_ERROR,
                original_exception=ConnectionError(
                    "Please ensured Docker is installed with BuildKit by following "
                    "https://docs.docker.com/build/buildkit/#getting-started"
                ),
            )

    def _build_and_tag(self, docker_config_dir: str) -> None:
        """Constructs the Docker context directory and then builds a Docker image based on that context.

        Args:
            docker_config_dir: Path to docker configuration directory, which stores the temporary session token.
        """
        self._build_image_from_context(docker_config_dir=docker_config_dir)

    def _run_docker_commands(self, commands: List[str]) -> None:
        """Run docker commands in a new child process.

        Args:
            commands: List of commands to run.

        Raises:
            SnowflakeMLException: Occurs when docker commands failed to execute.
        """
        proc = subprocess.Popen(
            commands, cwd=os.getcwd(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, shell=False
        )
        output_lines = []

        if proc.stdout:
            for line in iter(proc.stdout.readline, ""):
                output_lines.append(line)
                logger.debug(line)

        if proc.wait():
            for line in output_lines:
                logger.error(line)

            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INTERNAL_DOCKER_ERROR,
                original_exception=RuntimeError(f"Docker command failed: {' '.join(commands)}"),
            )

    def _build_image_from_context(self, docker_config_dir: str, *, platform: Platform = Platform.LINUX_AMD64) -> None:
        """Builds a Docker image based on provided context.

        Args:
            docker_config_dir: Path to docker configuration directory, which stores the temporary session token.
            platform: Target platform for the build output, in the format "os[/arch[/variant]]".
        """

        commands = [
            "docker",
            "--config",
            docker_config_dir,
            "buildx",
            "build",
            "--platform",
            platform.value,
            "--tag",
            f"{self.full_image_name}",
            self.context_dir,
        ]

        self._run_docker_commands(commands)

    def _upload(self, docker_config_dir: str) -> None:
        """
        Uploads image to the image registry. This process requires a "docker login" followed by a "docker push". Remove
        local image at the end of the upload operation to save up local space. Image cache is kept for more performant
        built experience at the cost of small storage footprint.

        By default, Docker overwrites the local Docker config file "/.docker/config.json" whenever a docker login
        occurs. However, to ensure better isolation between Snowflake-managed Docker credentials and the user's own
        Docker credentials, we will not use the default Docker config. Instead, we will write the username and session
        token to a temporary file and use "docker --config" so that it only applies to the specific Docker command being
        executed, without affecting the user's local Docker setup. The credential file will be automatically removed
        at the end of upload operation.

        Args:
            docker_config_dir: Path to docker configuration directory, which stores the temporary session token.
        """
        commands = ["docker", "--config", docker_config_dir, "login", self.full_image_name]
        self._run_docker_commands(commands)

        logger.debug(f"Pushing image to image repo {self.full_image_name}")
        commands = ["docker", "--config", docker_config_dir, "push", self.full_image_name]
        self._run_docker_commands(commands)



import inspect
import logging
import time
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import fsspec
from fsspec.implementations import http as httpfs

from snowflake import snowpark
from snowflake.connector import connection
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
    fileset_error_messages,
    fileset_errors,
)
from snowflake.snowpark import exceptions as snowpark_exceptions

# The default length of how long a presigned url stays active in seconds.
# Presigned url here is used to fetch file objects from Snowflake when SFStageFileSystem.open() is called.
_PRESIGNED_URL_LIFETIME_SEC = 14400

# The threshold of when the presigned url should get refreshed before its expiration.
_PRESIGNED_URL_HEADROOM_SEC = 3600


_PROJECT = "FileSet"


@dataclass(frozen=True)
class _PresignedUrl:
    """File system metadata for each stage file."""

    __slots__ = ["url", "expire_at"]
    url: str
    expire_at: float

    def is_expiring(self, headroom_sec: float = _PRESIGNED_URL_HEADROOM_SEC) -> bool:
        """Check if a token is going to expire in <headroom_sec> seconds."""
        return not self.expire_at or time.time() > self.expire_at - headroom_sec


def _get_httpfs_kwargs(**kwargs: Any) -> Dict[str, Any]:
    """Extract kwargs that are meaningful to HTTPFileSystem."""
    httpfs_related_keys = [
        "block_size",
        "cache_type",
        "cache_options",
    ]
    httpfs_related_kwargs = {}
    for key in httpfs_related_keys:
        v = kwargs.get(key, None)
        if v:
            httpfs_related_kwargs[key] = v
    return httpfs_related_kwargs


class SFStageFileSystem(fsspec.AbstractFileSystem):
    """ A Snowflake stage file system based on fsspec (https://filesystem-spec.readthedocs.io/).

    It grants user readonly access to Snowflake stage as if it were a file system. It exposes a
    filesystem-like API (ls, open) on top of Snowflake internal stage storage.

    Example: Create FS object and do file operation
    --------
    >>> conn = snowflake.connector.connect(**connection_parameters)
    >>> sffs = SFStageFileSystem(db="MYDB", schema="public", stage="FOO", sf_connection=conn)
    >>> sffs.ls("nytrain")
    ['nytrain/data_0_0_0.csv', 'nytrain/data_0_0_1.csv']
    >>> with sffs.open('nytrain/data_0_0_1.csv', mode='rb') as f:
    >>>     print(f.readline())
    b'2014-02-05 14:35:00.00000054,13,2014-02-05 14:35:00 UTC,\
        -74.00688,40.73049,-74.00563,40.70676,2\n'
    """

    def __init__(
        self,
        *,
        db: str,
        schema: str,
        stage: str,
        snowpark_session: Optional[snowpark.Session] = None,
        sf_connection: Optional[connection.SnowflakeConnection] = None,
        **kwargs: Any,
    ) -> None:
        """Initiate the file system with stage information and a snowflake connection.

        Args:
            db: The database name of the target internal stage.
            schema: The schema name of the target internal stage.
            stage: The name of the target stage.
            snowpark_session: A Snowpark session object. Mutually exclusive to `sf_connection`.
            sf_connection: A Snowflake python connection object. Mutually exclusive to `snowpark_session`.
            **kwargs : Optional. Other parameters that can be passed on to fsspec. Currently supports:
                - skip_instance_cache: Int. Controls reuse of instances.
                - cache_type, cache_options, block_size: Configure file buffering.
                See more information in https://filesystem-spec.readthedocs.io/en/latest/features.html

        Raises:
            ValueError: An error occurred when not exactly one of sf_connection and snowpark_session is given.
        """
        if sf_connection and snowpark_session:
            raise ValueError(fileset_error_messages.BOTH_SF_CONNECTION_AND_SNOWPARK_SESSION_SPECIFIED)
        if not sf_connection and not snowpark_session:
            raise ValueError(fileset_error_messages.NO_SF_CONNECTION_OR_SNOWPARK_SESSION)
        if sf_connection:
            self._session = snowpark.Session.builder.config("connection", sf_connection).create()
        else:
            self._session = snowpark_session

        logging.debug(f"Creating new stage file system on @{db}.{schema}.{stage}")
        self._db = db
        self._schema = schema
        self._stage = stage
        self._url_cache: Dict[str, _PresignedUrl] = {}

        httpfs_kwargs = _get_httpfs_kwargs(**kwargs)
        self._fs = httpfs.HTTPFileSystem(**httpfs_kwargs)

        super().__init__(**kwargs)

    @property
    def stage_name(self) -> str:
        """Get the Snowflake path to this stage.

        Returns:
            String in the format of "@{database}.{schema}.{stage}".
                Example: @mydb.myschema.mystage
        """
        return f"@{self._db}.{self._schema}.{self._stage}"

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
        func_params_to_log=["detail"],
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    def ls(self, path: str, detail: bool = False) -> Union[List[str], List[Dict[str, Any]]]:
        """Override fsspec `ls` method. List single "directory" with or without details.

        Args:
            path: Relative file paths in the stage.
                Example:
                    "": Empty string points to the stage root.
                    "mydir": Points to the "mydir" file/directory under the stage.
                    "dir1/dir2": Points to the "dir2" file/directory in the "dir1" directory.
            detail: Whether to present detailed information of results. If set to be False, a list of filenames will be
                returned. If set to be True, each list item in the result is a dict, whose keys are "name", "size",
                "type", "md5" and "last_modified".

        Returns:
            A list of filename if `detail` is false, or a list of dict if `detail` is true.

        Raises:
            SnowflakeMLException: An error occurred when the given path points to a stage that cannot be found.
            SnowflakeMLException: An error occurred when Snowflake cannot list files in the given stage path.
        """
        try:
            loc = self.stage_name
            objects = self._session.sql(f"LIST {loc}/{path}").collect()
        except snowpark_exceptions.SnowparkClientException as e:
            if e.message.startswith(fileset_errors.ERRNO_DOMAIN_NOT_EXIST):
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.SNOWML_NOT_FOUND,
                    original_exception=fileset_errors.StageNotFoundError(
                        f"Stage {loc} does not exist or is not authorized."
                    ),
                )
            else:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INTERNAL_SNOWML_ERROR,
                    original_exception=fileset_errors.FileSetError(str(e)),
                )
        files = self._parse_list_result(objects, path)
        if detail:
            return files
        else:
            return [f["name"] for f in files]

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    def optimize_read(self, files: Optional[List[str]] = None) -> None:
        """Prefetch and cache the presigned urls for all the given files to speed up the read performance.

        All the files introduced here will have their urls cached. Further open() on any of cached urls will lead to a
        batch refreshment of all the cached urls if that url is inactive.

        Args:
            files: A list of file paths. If not given, all the files that have urls cached will refresh their url cache.
        """
        if not files:
            files = list(self._url_cache.keys())
        url_lifetime = _PRESIGNED_URL_LIFETIME_SEC
        start_time = time.time()
        logging.info(f"Start batch fetching presigned urls for {self.stage_name}.")
        presigned_urls = self._fetch_presigned_urls(files, url_lifetime)
        expire_at = start_time + url_lifetime

        for presigned_url in presigned_urls:
            file_path, url = presigned_url
            self._url_cache[file_path] = _PresignedUrl(url, expire_at)
            logging.debug(f"Retrieved presigned url for {file_path}.")
        logging.info(f"Finished batch fetching presigned urls for {self.stage_name}.")

    @telemetry.send_api_usage_telemetry(
        project=_PROJECT,
    )
    @snowpark._internal.utils.private_preview(version="0.2.0")
    def _open(self, path: str, mode: str = "rb", **kwargs: Any) -> fsspec.spec.AbstractBufferedFile:
        """Override fsspec `_open` method. Open a file for reading.

        The opened file will be readable for 4 hours. After that, you need to reopen the file.

        Args:
            path: Path of file in Snowflake stage.
            mode: One of 'r', 'rb'. These have the same meaning as they do for the built-in `open` function.
            **kwargs: Extra options that supported by fsspec. See more in
                https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.open

        Returns:
            A fsspec file-like object.

        Raises:
            SnowflakeMLException: An error occurred when the given path points to a file that cannot be found.
        """
        path = path.lstrip("/")
        cached_presigned_url = self._url_cache.get(path, None)
        if not cached_presigned_url:
            res = self._fetch_presigned_urls([path])
            url = res[0][1]
            expire_at = time.time() + _PRESIGNED_URL_LIFETIME_SEC
            cached_presigned_url = _PresignedUrl(url, expire_at)
            self._url_cache[path] = cached_presigned_url
            logging.debug(f"Retrieved presigned url for {path}.")
        elif cached_presigned_url.is_expiring():
            self.optimize_read()
            cached_presigned_url = self._url_cache[path]
        url = cached_presigned_url.url
        try:
            return self._fs._open(url, mode=mode, **kwargs)
        except FileNotFoundError:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.SNOWML_NOT_FOUND,
                original_exception=fileset_errors.StageFileNotFoundError(f"Stage file {path} doesn't exist."),
            )

    def _parse_list_result(
        self, list_result: List[Tuple[str, int, str, str]], search_path: str
    ) -> List[Dict[str, Any]]:
        """Convert the result from LIST query to the expected format of fsspec ls() method.

        Note that Snowflake LIST query has different behavior with ls(). LIST query will return all the stage files
        whose path starts with the given search_path. However, ls() should only return files that either exactly
        match the given search_path or are under the given search_path directory.

        For example, both "mydir/hello_world.txt" and "mydir1/hello_world.txt" will appear as a result of "LIST mydir",
        but only the former one is in "mydir" directory. As a result, only "mydir/hello_world.txt" should be returned by
        ls("mydir").

        Args:
            list_result: The result of LIST query, a list where each item is a Snowpark Row with four items:
                name, size, md5 and last_modified.
            search_path: The path that was used by the List query to get the list_result.

        Returns:
            A list of dict, where each dict contains key-value pairs as the properties of a file.
        """
        files: Dict[str, Dict[str, Any]] = {}
        search_path = search_path.strip("/")
        for name, size, md5, last_modified in list_result:
            obj_path = self._stage_path_to_relative_path(name)
            if obj_path == search_path:
                # If there is a exact match, then the matched object will always be a file object.
                self._add_file_info_helper(files, obj_path, size, "file", md5, last_modified)
                continue
            elif search_path and not obj_path.startswith(search_path + "/"):
                # If the path doesn't start with "<search_path>/", the object is not under the <search_path> directory.
                continue

            # Now we want to distinguish whether the object is in a subdirecotry of the given search path.
            rel_file_path = obj_path[len(search_path) :].lstrip("/")
            slash_idx = rel_file_path.find("/")
            if slash_idx == -1:
                # There is no subdirectory.
                self._add_file_info_helper(files, obj_path, size, "file", md5, last_modified)
            else:
                # There is a subdirectory. We will only add the top level subdirectory to the result of ls().
                dir_path = "" if not search_path else f"{search_path}/"
                dir_path += f"{rel_file_path[:slash_idx + 1]}"
                self._add_file_info_helper(files, dir_path, 0, "directory", None, None)
        return list(files.values())

    def _stage_path_to_relative_path(self, stage_path: str) -> str:
        """Convert a stage file path which comes from the LIST query to a relative file path in that stage.

        The file path returned by LIST query always has the format "<stage_name>/<relative_file_path>".
                Only <relative_file_path> will be returned.

        Args:
            stage_path: A string started with the name of the stage.

        Returns:
            A string of the relative stage path.
        """
        return stage_path[len(self._stage) + 1 :]

    def _add_file_info_helper(
        self,
        files: Dict[str, Dict[str, Any]],
        object_path: str,
        file_size: int,
        file_type: str,
        md5: Optional[str],
        last_modified: Optional[str],
    ) -> None:
        files.setdefault(
            object_path,
            {
                "name": object_path,
                "size": file_size,
                "type": file_type,
                "md5": md5,
                "last_modified": last_modified,
            },
        )

    def _fetch_presigned_urls(
        self, files: List[str], url_lifetime: float = _PRESIGNED_URL_LIFETIME_SEC
    ) -> List[Tuple[str, str]]:
        """Fetch presigned urls for the given files."""
        file_df = self._session.create_dataframe(files).to_df("name")
        try:
            presigned_urls: List[Tuple[str, str]] = file_df.select_expr(
                f"name, get_presigned_url({self.stage_name}, name, {url_lifetime}) as url"
            ).collect(
                statement_params=telemetry.get_function_usage_statement_params(
                    project=_PROJECT,
                    function_name=telemetry.get_statement_params_full_func_name(
                        inspect.currentframe(), self.__class__.__name__
                    ),
                    api_calls=[snowpark.DataFrame.collect],
                ),
            )
        except snowpark_exceptions.SnowparkClientException as e:
            if e.message.startswith(fileset_errors.ERRNO_STAGE_NOT_EXIST):
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.SNOWML_NOT_FOUND,
                    original_exception=fileset_errors.StageNotFoundError(
                        f"Stage {self.stage_name} does not exist or is not authorized."
                    ),
                )
            else:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INTERNAL_SNOWML_ERROR,
                    original_exception=fileset_errors.FileSetError(str(e)),
                )
        return presigned_urls



import os
from types import ModuleType
from typing import Dict, List, Optional

from absl import logging

from snowflake.ml._internal import env_utils
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml.model import custom_model, model_signature, type_hints as model_types
from snowflake.ml.model._packager import model_handler
from snowflake.ml.model._packager.model_meta import model_meta


class ModelPackager:
    """Top-level class to save/load and manage a Snowflake Native formatted model.
        It maintains the actual model blob files, environment required by model itself and signatures to do the
        inference with the model.

    Attributes:
        local_dir_path: A path to a local directory will files to dump and load.
        model: The model object to be saved / loaded from file.
        meta: The model metadata (ModelMetadata object) to be saved / loaded from file.
            model and meta will be set once save / load method is called.

    """

    MODEL_BLOBS_DIR = "models"

    def __init__(self, local_dir_path: str) -> None:
        self.local_dir_path = os.path.normpath(local_dir_path)
        self.model: Optional[model_types.SupportedModelType] = None
        self.meta: Optional[model_meta.ModelMetadata] = None

    def save(
        self,
        *,
        name: str,
        model: model_types.SupportedModelType,
        signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
        sample_input: Optional[model_types.SupportedDataType] = None,
        metadata: Optional[Dict[str, str]] = None,
        conda_dependencies: Optional[List[str]] = None,
        pip_requirements: Optional[List[str]] = None,
        python_version: Optional[str] = None,
        ext_modules: Optional[List[ModuleType]] = None,
        code_paths: Optional[List[str]] = None,
        options: Optional[model_types.ModelSaveOption] = None,
    ) -> None:
        if (signatures is None) and (sample_input is None) and not model_handler.is_auto_signature_model(model):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    "Signatures and sample_input both cannot be None at the same time for this kind of model."
                ),
            )

        if (signatures is not None) and (sample_input is not None):
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError("Signatures and sample_input both cannot be specified at the same time."),
            )

        if not options:
            options = model_types.BaseModelSaveOption()

        handler = model_handler.find_handler(model)
        if handler is None:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_TYPE,
                original_exception=TypeError(f"{type(model)} is not supported."),
            )
        with model_meta.create_model_metadata(
            model_dir_path=self.local_dir_path,
            name=name,
            model_type=handler.HANDLER_TYPE,
            metadata=metadata,
            code_paths=code_paths,
            signatures=signatures,
            ext_modules=ext_modules,
            conda_dependencies=conda_dependencies,
            pip_requirements=pip_requirements,
            python_version=python_version,
            **options,
        ) as meta:
            model_blobs_path = os.path.join(self.local_dir_path, ModelPackager.MODEL_BLOBS_DIR)
            os.makedirs(model_blobs_path, exist_ok=True)
            model = handler.cast_model(model)
            handler.save_model(
                name=name,
                model=model,
                model_meta=meta,
                model_blobs_dir_path=model_blobs_path,
                sample_input=sample_input,
                is_sub_model=False,
                **options,
            )
            if signatures is None:
                logging.info(f"Model signatures are auto inferred as:\n\n{meta.signatures}")

            self.model = model
            self.meta = meta

    def load(
        self,
        *,
        meta_only: bool = False,
        as_custom_model: bool = False,
        options: Optional[model_types.ModelLoadOption] = None,
    ) -> None:
        """Load the model into memory from directory. Used internal only.

        Args:
            meta_only: Flag to indicate that if only load metadata.
            as_custom_model: When set to True, It will try to convert the model as custom model after load.
            options: Model loading options.

        Raises:
            SnowflakeMLException: Raised if model is not native format.
        """

        self.meta = model_meta.ModelMetadata.load(self.local_dir_path)
        if meta_only:
            return

        model_meta.load_code_path(self.local_dir_path)

        env_utils.validate_py_runtime_version(self.meta.env.python_version)

        handler = model_handler.load_handler(self.meta.model_type)
        if handler is None:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_TYPE,
                original_exception=TypeError(f"{self.meta.model_type} is not supported."),
            )
        model_blobs_path = os.path.join(self.local_dir_path, ModelPackager.MODEL_BLOBS_DIR)
        if options is None:
            options = {}

        handler.try_upgrade(self.meta.name, self.meta, model_blobs_path)
        m = handler.load_model(self.meta.name, self.meta, model_blobs_path, **options)

        if as_custom_model:
            m = handler.convert_as_custom_model(m, self.meta, **options)
            assert isinstance(m, custom_model.CustomModel)

        self.model = m



import functools
import inspect
from typing import Any, Callable, Coroutine, Dict, Generator, Optional

import anyio
import pandas as pd

from snowflake.ml.model import type_hints as model_types


class MethodRef:
    """Represents an method invocation of an instance of `ModelRef`.

    This allows us to
        1) Customize the place of actual execution of the method(inline, thread/process pool or remote).
        2) Enrich the way of execution(sync versus async).
    Example:
        If you have a SKL model, you would normally invoke by `skl_ref.predict(df)` which has sync API.
        Within inference graph, you could invoke `await skl_ref.predict.async_run(df)` which automatically
        will be run on thread with async interface.
    """

    def __init__(self, model_ref: "ModelRef", method_name: str) -> None:
        self._func = getattr(model_ref.model, method_name)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._func(*args, **kwargs)

    async def async_run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the method in a async way. If the method is defined as async, this will simply run it. If not, this will
        be run in a separate thread.

        Args:
            *args: Arguments of the original method,
            **kwargs: Keyword arguments of the original method.

        Returns:
            The result of the original method.
        """
        if inspect.iscoroutinefunction(self._func):
            return await self._func(*args, **kwargs)
        return await anyio.to_thread.run_sync(functools.partial(self._func, **kwargs), *args)


class ModelRef:
    """Represents an model in the inference graph. Method could be directly called using this reference object as if
    with the original model object.

    This enables us to separate physical and logical representation of a model which
    will allows us to deeply understand the graph and perform optimization at entire
    graph level.
    """

    def __init__(self, name: str, model: model_types.SupportedModelType) -> None:
        """Initialize the ModelRef.

        Args:
            name: The name of a model to refer it.
            model: The model object.
        """
        self._model = model
        self._name = name

    @property
    def name(self) -> str:
        """The name of the sub-model."""
        return self._name

    @property
    def model(self) -> model_types.SupportedModelType:
        """The model object of the sub-model."""
        return self._model

    def __getattr__(self, method_name: str) -> Any:
        if hasattr(self._model, method_name):
            return MethodRef(self, method_name)
        raise TypeError(f"Model is does not have {method_name}.")

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        del state["_model"]
        return state

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        if callable(self._model):
            return MethodRef(self, "__call__")(*args, **kwds)
        raise TypeError("Model is not callable.")

    def __setstate__(self, state: Any) -> None:
        self.__dict__.update(state)


class ModelContext:
    """Context for a custom model showing path to artifacts and mapping between model name and object reference.

    Attributes:
        artifacts: A dict mapping name of the artifact to its path.
        model_refs: A dict mapping name of the sub-model to its ModelRef object.
    """

    def __init__(
        self,
        *,
        artifacts: Optional[Dict[str, str]] = None,
        models: Optional[Dict[str, model_types.SupportedModelType]] = None,
    ) -> None:
        """Initialize the model context

        Args:
            artifacts: A dict mapping name of the artifact to its currently available path. Defaults to None.
            models: A dict mapping name of the sub-model to the corresponding model object. Defaults to None.
        """
        self.artifacts: Dict[str, str] = artifacts if artifacts else dict()
        self.model_refs: Dict[str, ModelRef] = (
            {name: ModelRef(name, model) for name, model in models.items()} if models else dict()
        )

    def path(self, key: str) -> str:
        """Get the actual path to a specific artifact.

        Args:
            key: The name of the artifact.

        Returns:
            The absolute path to the artifact.
        """
        return self.artifacts[key]

    def model_ref(self, name: str) -> ModelRef:
        """Get a ModelRef object of a sub-model containing the name and model object, while able to call its method
        directly as well.

        Args:
            name: The name of the sub-model.

        Returns:
            The ModelRef object to the sub-model.
        """
        return self.model_refs[name]


class CustomModel:
    """Abstract class for user defined custom model.

    Attributes:
        context: A ModelContext object showing sub-models and artifacts related to this model.
    """

    def __init__(self, context: ModelContext) -> None:
        self.context = context
        for method in self._get_infer_methods():
            _validate_predict_function(method)

    def __setattr__(self, __name: str, __value: Any) -> None:
        # A hook for case when users reassign the method.
        if getattr(__value, "_is_inference_api", False):
            if inspect.ismethod(__value):
                _validate_predict_function(__value.__func__)
            else:
                raise TypeError("A non-method inference API function is not supported.")
        super().__setattr__(__name, __value)

    def _get_infer_methods(
        self,
    ) -> Generator[Callable[[model_types.CustomModelType, pd.DataFrame], pd.DataFrame], None, None]:
        """Returns all methods in CLS with DECORATOR as the outermost decorator."""
        for cls_method_str in dir(self):
            cls_method = getattr(self, cls_method_str)
            if getattr(cls_method, "_is_inference_api", False):
                if inspect.ismethod(cls_method):
                    yield cls_method.__func__
                else:
                    raise TypeError("A non-method inference API function is not supported.")


def _validate_predict_function(func: Callable[[model_types.CustomModelType, pd.DataFrame], pd.DataFrame]) -> None:
    """Validate the user provided predict method.

    Args:
        func: The predict method.

    Raises:
        TypeError: Raised when the method is not a callable object.
        TypeError: Raised when the method does not have 2 arguments (self and X).
        TypeError: Raised when the method does not have typing annotation.
        TypeError: Raised when the method's input (X) does not have type pd.DataFrame.
        TypeError: Raised when the method's output does not have type pd.DataFrame.
    """
    if not callable(func):
        raise TypeError("Predict method is not callable.")

    func_signature = inspect.signature(func)
    if len(func_signature.parameters) != 2:
        raise TypeError("Predict method should have exact 2 arguments.")

    input_annotation = list(func_signature.parameters.values())[1].annotation
    output_annotation = func_signature.return_annotation

    if input_annotation == inspect.Parameter.empty or output_annotation == inspect.Signature.empty:
        raise TypeError("Missing type annotation for predict method.")

    if input_annotation != pd.core.frame.DataFrame:
        raise TypeError("Input for predict method should have type pandas.DataFrame.")

    if (
        output_annotation != pd.core.frame.DataFrame
        and output_annotation != Coroutine[Any, Any, pd.core.frame.DataFrame]
    ):
        raise TypeError("Output for predict method should have type pandas.DataFrame.")


def inference_api(
    func: Callable[[model_types.CustomModelType, pd.DataFrame], pd.DataFrame]
) -> Callable[[model_types.CustomModelType, pd.DataFrame], pd.DataFrame]:
    func.__dict__["_is_inference_api"] = True
    return func
